from spaceone.core import cache
from spaceone.core.service import *
from spaceone.core import utils
from spaceone.identity.manager.provider_manager import ProviderManager


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class ProviderService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.provider_mgr: ProviderManager = self.locator.get_manager('ProviderManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['provider', 'name', 'domain_id'])
    def create(self, params):
        """
        Args:
            params (dict): {
                'provider': 'str',
                'name': 'str',
                'template': 'dict',
                'metadata': 'dict',
                'capability': 'dict',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            provider_vo (object)
        """
        # TODO: validate a template data
        # TODO: validate a capability data

        return self.provider_mgr.create_provider(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['provider', 'domain_id'])
    def update(self, params):
        """
        Args:
            params (dict): {
                'provider': 'str',
                'name': 'str',
                'template': 'dict',
                'metadata': 'dict',
                'capability': 'dict',
                'tags': 'list',
                'domain_id': 'str'
            }

        Returns:
            provider_vo (object)
        """
        # TODO: validate a template data
        # TODO: validate a capability data

        return self.provider_mgr.update_provider(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['provider', 'domain_id'])
    def delete(self, params):
        """
        Args:
            params (dict): {
                'provider': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """
        self.provider_mgr.delete_provider(params['provider'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['provider', 'domain_id'])
    def get(self, params):
        """
        Args:
            params (dict): {
                'provider': 'str',
                'only': 'list',
                'domain_id': 'str'
            }

        Returns:
            provider_vo (object)
        """
        self._create_default_provider()
        return self.provider_mgr.get_provider(params['provider'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @append_query_filter(['provider', 'name'])
    @append_keyword_filter(['provider', 'name'])
    def list(self, params):
        """
        Args:
            params (dict): {
                    'query': 'dict (spaceone.api.core.v1.Query)',
                    'provider': 'str',
                    'name': 'str',
                    'domain_id': 'str'
                }

        Returns:
            results (list): 'list of provider_vo'
            total_count (int)
        """
        self._create_default_provider()
        return self.provider_mgr.list_providers(params.get('query', {}))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query', 'domain_id'])
    @append_keyword_filter(['provider', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'domain_id': 'str'
            }

        Returns:
            values (list): 'list of statistics data'
            total_count (int)
        """

        query = params.get('query', {})
        return self.provider_mgr.stat_providers(query)

    @cache.cacheable(key='provider:default:init', expire=300)
    def _create_default_provider(self):
        provider_vos, total_count = self.provider_mgr.list_providers()
        installed_providers = [provider_vo.provider for provider_vo in provider_vos]
        self.provider_mgr.create_default_providers(installed_providers)

        return True



from mongoengine import *
from spaceone.core.model.mongo_model import MongoModel


class User(MongoModel):
    user_id = StringField(max_length=40, unique_with='domain_id', required=True)
    password = BinaryField(default=None)
    name = StringField(max_length=128, default='')
    state = StringField(max_length=20, choices=('ENABLED', 'DISABLED', 'PENDING'))
    email = StringField(max_length=255, default='')
    email_verified = BooleanField(default=False)
    user_type = StringField(max_length=20, choices=('USER', 'API_USER'))
    backend = StringField(max_length=20, choices=('LOCAL', 'EXTERNAL'))
    required_actions = ListField(StringField(choices=('UPDATE_PASSWORD',)), default=[])
    language = StringField(max_length=7, default='en')
    timezone = StringField(max_length=50, default='UTC')
    tags = DictField()
    domain_id = StringField(max_length=40)
    last_accessed_at = DateTimeField(default=None, null=True)
    created_at = DateTimeField(auto_now_add=True)

    meta = {
        'updatable_fields': [
            'password',
            'name',
            'state',
            'email',
            'email_verified',
            'required_actions',
            'language',
            'timezone',
            'tags',
            'last_accessed_at'
        ],
        'minimal_fields': [
            'user_id',
            'name',
            'state',
            'user_type'
        ],
        'ordering': ['name'],
        'indexes': [
            'state',
            'user_type',
            'backend',
            'last_accessed_at',
            # ('user_id', 'domain_id'),
        ]
    }



import logging
from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.identity.error.error_role import *
from spaceone.identity.model.role_binding_model import *
from spaceone.identity.manager import RoleManager, ProjectManager, ProjectGroupManager, UserManager

_LOGGER = logging.getLogger(__name__)
_SUPPORTED_RESOURCE_TYPES = ['identity.User']


class RoleBindingManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.role_binding_model: RoleBinding = self.locator.get_model('RoleBinding')

    def create_role_binding(self, params):
        def _rollback(role_binding_vo):
            _LOGGER.info(f'[create_role_binding._rollback] Delete role binding : {role_binding_vo.name} ({role_binding_vo.role_binding_id})')
            role_binding_vo.delete()

        resource_type = params['resource_type']
        resource_id = params['resource_id']
        role_id = params['role_id']
        project_id = params.get('project_id')
        project_group_id = params.get('project_group_id')
        domain_id = params['domain_id']

        role_mgr: RoleManager = self.locator.get_manager('RoleManager')
        project_mgr: ProjectManager = self.locator.get_manager('ProjectManager')
        project_group_mgr: ProjectGroupManager = self.locator.get_manager('ProjectGroupManager')
        user_mgr: UserManager = self.locator.get_manager('UserManager')

        self._check_resource_type(resource_type)
        if resource_type == 'identity.User':
            params['user'] = user_mgr.get_user(resource_id, domain_id)

        role_vo = role_mgr.get_role(role_id, domain_id)
        self._check_role_type(role_vo.role_type, resource_type, resource_id, domain_id)
        params['role'] = role_vo

        if role_vo.role_type == 'PROJECT':
            if project_id:
                project_vo = project_mgr.get_project(project_id, domain_id)
                self._check_duplicate_project_role(resource_type, resource_id, project_vo, project_id)
                params['project'] = project_vo

            elif project_group_id:
                project_group_vo = project_group_mgr.get_project_group(project_group_id, domain_id)
                self._check_duplicate_project_group_role(resource_type, resource_id, project_group_vo, project_group_id)
                params['project_group'] = project_group_vo
            else:
                raise ERROR_REQUIRED_PROJECT_OR_PROJECT_GROUP()
        else:
            self._check_duplicate_domain_and_system_role(resource_type, resource_id, role_vo, role_id)
            if project_id:
                raise ERROR_NOT_ALLOWED_PROJECT_ID()
            elif project_group_id:
                raise ERROR_NOT_ALLOWED_PROJECT_GROUP_ID()

        role_binding_vo: RoleBinding = self.role_binding_model.create(params)
        self.transaction.add_rollback(_rollback, role_binding_vo)

        if role_vo.role_type in ['DOMAIN', 'SYSTEM']:
            self._delete_old_domain_or_system_role_binding(resource_id, resource_type, role_vo.role_type,
                                                           role_binding_vo.role_binding_id, domain_id)

        cache.delete_pattern(f'role-bindings:{domain_id}:{resource_id}*')
        cache.delete_pattern(f'user-permissions:{domain_id}:{resource_id}*')
        cache.delete_pattern(f'user-scopes:{domain_id}:{resource_id}*')

        return role_binding_vo

    def update_role_binding(self, params):
        role_binding_vo = self.get_role_binding(params['role_binding_id'], params['domain_id'])
        return self.update_role_binding_by_vo(params, role_binding_vo)

    def update_role_binding_by_vo(self, params, role_binding_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_role_binding._rollback] Revert Data : {old_data["role_binding_id"]}')
            role_binding_vo.update(old_data)

        self.transaction.add_rollback(_rollback, role_binding_vo.to_dict())

        return role_binding_vo.update(params)

    def delete_role_binding(self, role_binding_id, domain_id):
        role_binding_vo = self.get_role_binding(role_binding_id, domain_id)
        self.delete_role_binding_by_vo(role_binding_vo)

    def delete_role_binding_by_vo(self, role_binding_vo):
        resource_id = role_binding_vo.resource_id
        domain_id = role_binding_vo.domain_id

        _LOGGER.debug(f'[delete_role_binding_by_vo] resource_id = {resource_id}, domain_id = {domain_id}')

        role_binding_vo.delete()

        cache.delete_pattern(f'role-bindings:{domain_id}:{resource_id}*')
        cache.delete_pattern(f'user-permissions:{domain_id}:{resource_id}*')
        cache.delete_pattern(f'user-scopes:{domain_id}:{resource_id}*')

    def get_role_binding(self, role_binding_id, domain_id, only=None):
        return self.role_binding_model.get(role_binding_id=role_binding_id, domain_id=domain_id, only=only)

    def get_project_role_binding(self, resource_type, resource_id, domain_id, project_vo=None, project_group_vo=None):
        return self.role_binding_model.filter(resource_type=resource_type, resource_id=resource_id, domain_id=domain_id,
                                              project=project_vo, project_group=project_group_vo)

    def get_user_role_bindings(self, user_id, domain_id):
        return self.role_binding_model.filter(resource_type='identity.User', resource_id=user_id, domain_id=domain_id)

    def list_role_bindings(self, query):
        return self.role_binding_model.query(**query)

    def stat_role_bindings(self, query):
        return self.role_binding_model.stat(**query)

    def _delete_old_domain_or_system_role_binding(self, resource_id, resource_type, role_type, new_role_binding_id,
                                                  domain_id):
        query = {
            'filter': [
                {'k': 'resource_id', 'v': resource_id, 'o': 'eq'},
                {'k': 'resource_type', 'v': resource_type, 'o': 'eq'},
                {'k': 'role.role_type', 'v': role_type, 'o': 'eq'},
                {'k': 'role_binding_id', 'v': new_role_binding_id, 'o': 'not'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'},
            ]
        }

        _LOGGER.debug(f'[_delete_old_domain_or_system_role_binding] query = {query}')

        rb_vos, total_count = self.list_role_bindings(query)
        rb_vos.delete()

    @staticmethod
    def _check_resource_type(resource_type):
        if resource_type not in _SUPPORTED_RESOURCE_TYPES:
            raise ERROR_INVALID_PARAMETER(
                key='resource_type', reason=f'resource_type is not supported. (support = {_SUPPORTED_RESOURCE_TYPES})')

    def _check_role_type(self, role_type, resource_type, resource_id, domain_id):
        role_binding_vos = self.role_binding_model.filter(resource_type=resource_type,
                                                          resource_id=resource_id, domain_id=domain_id)
        for role_binding_vo in role_binding_vos:
            if role_type == 'SYSTEM':
                if role_binding_vo.role.role_type in ['PROJECT', 'DOMAIN']:
                    raise ERROR_NOT_ALLOWED_ROLE_TYPE()
            else:
                if role_binding_vo.role.role_type == 'SYSTEM':
                    raise ERROR_NOT_ALLOWED_ROLE_TYPE()

    def _check_duplicate_domain_and_system_role(self, resource_type, resource_id, role_vo, role_id):
        rb_vos = self.role_binding_model.filter(resource_type=resource_type, resource_id=resource_id, role=role_vo)

        if rb_vos.count() > 0:
            raise ERROR_DUPLICATE_ROLE_BOUND(role_id=role_id, resource_id=resource_id)

    def _check_duplicate_project_role(self, resource_type, resource_id, project_vo, project_id):
        project_rb_vos = self.role_binding_model.filter(resource_type=resource_type, resource_id=resource_id,
                                                        project=project_vo)

        if project_rb_vos.count() > 0:
            raise ERROR_DUPLICATE_RESOURCE_IN_PROJECT(project_id=project_id, resource_id=resource_id)

    def _check_duplicate_project_group_role(self, resource_type, resource_id, project_group_vo, project_group_id):
        pg_rb_vos = self.role_binding_model.filter(resource_type=resource_type, resource_id=resource_id,
                                                   project_group=project_group_vo)

        if pg_rb_vos.count() > 0:
            raise ERROR_DUPLICATE_RESOURCE_IN_PROJECT_GROUP(project_group_id=project_group_id,
                                                            resource_id=resource_id)



from spaceone.core.service import *
from spaceone.core import utils
from spaceone.identity.manager import RoleBindingManager


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class RoleBindingService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.role_binding_mgr: RoleBindingManager = self.locator.get_manager('RoleBindingManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['resource_type', 'resource_id', 'role_id', 'domain_id'])
    def create(self, params):
        """ Create role binding

        Args:
            params (dict): {
                'resource_type': 'str',
                'resource_id': 'str',
                'role_id': 'str',
                'project_id': 'str',
                'project_group_id': 'str',
                'labels': 'list',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            role_binding_vo (object)
        """

        return self.role_binding_mgr.create_role_binding(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['role_binding_id', 'domain_id'])
    def update(self, params):
        """ Update role binding

        Args:
            params (dict): {
                'role_binding_id': 'str',
                'labels': 'list',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            role_binding_vo (object)
        """

        return self.role_binding_mgr.update_role_binding(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['role_binding_id', 'domain_id'])
    def delete(self, params):
        """ Delete role binding

        Args:
            params (dict): {
                'role_binding_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.role_binding_mgr.delete_role_binding(params['role_binding_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['role_binding_id', 'domain_id'])
    @change_only_key({'project_group_info': 'project_group', 'project_info': 'project', 'role_info': 'role'})
    def get(self, params):
        """ Get role binding

        Args:
            params (dict): {
                'role_binding_id': 'str',
                'domain_id': 'str',
                'only': 'list'
            }

        Returns:
            role_binding_vo (object)
        """

        return self.role_binding_mgr.get_role_binding(params['role_binding_id'], params['domain_id'],
                                                      params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @change_only_key({'project_group_info': 'project_group', 'project_info': 'project', 'role_info': 'role'},
                     key_path='query.only')
    @append_query_filter(['role_binding_id', 'resource_type', 'resource_id', 'role_id', 'role_type',
                          'project_id', 'project_group_id', 'domain_id'])
    @append_keyword_filter(['role_binding_id', 'resource_id', 'name'])
    def list(self, params):
        """ List role bindings

        Args:
            params (dict): {
                'role_binding_id': 'str',
                'resource_type': 'str',
                'resource_id': 'str',
                'role_id': 'str',
                'role_type': 'str',
                'project_id': 'str',
                'project_group_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of role_binding_vo'
            total_count (int)
        """
        query = params.get('query', {})

        return self.role_binding_mgr.list_role_bindings(params.get('query', {}))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @append_keyword_filter(['role_binding_id', 'resource_type', 'resource_id'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list): 'list of statistics data'
            total_count (int)
        """

        query = params.get('query', {})
        return self.role_binding_mgr.stat_role_bindings(query)



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel
from spaceone.identity.model.domain_model import Domain


class DomainSecret(MongoModel):
    domain_key = StringField()
    pub_jwk = DictField(required=True)
    prv_jwk = DictField(required=True)
    refresh_pub_jwk = DictField(required=True)
    refresh_prv_jwk = DictField(required=True)
    domain_id = StringField(max_length=40, unique=True)
    domain = ReferenceField('Domain', reverse_delete_rule=CASCADE)
    created_at = DateTimeField(auto_now_add=True)

    meta = {
        'updatable_fields': [
            'refresh_pub_jwk',
            'refresh_prv_jwk',
            'domain'
        ],
        'ordering': ['domain_id'],
        'indexes': [
            # 'domain_id',
            'domain'
        ]
    }



import fnmatch
import functools
import logging

from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.core.error import *

_LOGGER = logging.getLogger(__name__)


class AuthorizationManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @cache.cacheable(key='user-permissions:{domain_id}:{user_id}:{service}:{resource}:{verb}:{request_roles}',
                     expire=3600)
    def check_permissions(self, user_id, domain_id, permissions, service, resource, verb, request_roles):
        try:
            next(filter(functools.partial(
                self._match_permission, service, resource, verb), permissions))
        except Exception as e:
            _LOGGER.debug(f'[verify] Not matched permissions. '
                          f'(user_id={user_id}, api={service}.{resource}.{verb})')
            raise ERROR_PERMISSION_DENIED()

    @cache.cacheable(key='user-scopes:{domain_id}:{user_id}:{scope}:{role_type}:{request_domain_id}:'
                         '{request_project_id}:{request_project_group_id}:{projects}:{project_groups}:'
                         '{request_user_id}:{require_project_id}:{require_project_group_id}:{require_user_id}:'
                         '{require_domain_id}',
                     expire=3600)
    def check_scope_by_role_type(self, user_id, domain_id, scope, role_type, projects, project_groups,
                                 request_domain_id, request_project_id, request_project_group_id, request_user_id,
                                 require_project_id, require_project_group_id, require_user_id, require_domain_id):

        if scope == 'SYSTEM':
            # Excluding system api checking process
            # self._check_system_scope(user_id, domain_id, role_type)
            pass
        elif scope == 'PUBLIC':
            # No Check Public Resource
            pass
        elif scope in ['DOMAIN', 'PUBLIC_OR_DOMAIN']:
            self._check_domain_scope(user_id, domain_id, role_type, request_domain_id, require_domain_id)
        elif scope in ['PROJECT', 'DOMAIN_OR_PROJECT']:
            self._check_domain_scope(user_id, domain_id, role_type, request_domain_id, require_domain_id)
            self._check_project_scope(user_id, domain_id, role_type, projects, project_groups,
                                      request_project_id, request_project_group_id,
                                      require_project_id, require_project_group_id)
        elif scope == 'USER':
            self._check_domain_scope(user_id, domain_id, role_type, request_domain_id, require_domain_id)
            self._check_user_scope(user_id, domain_id, role_type, request_user_id, require_user_id)

    @staticmethod
    def _check_system_scope(user_id, domain_id, role_type):
        if role_type != 'SYSTEM':
            _LOGGER.debug(f'[_check_system_scope] system api is not allowed.'
                          f' (user_id = {user_id}, user_domain_id = {domain_id})')
            raise ERROR_PERMISSION_DENIED()

    @staticmethod
    def _check_domain_scope(user_id, domain_id, role_type, request_domain_id, require_domain_id):
        if role_type != 'SYSTEM':
            if require_domain_id and request_domain_id is None:
                _LOGGER.debug(f'[_check_domain_scope] domain_id is required.'
                              f' (user_id = {user_id}, user_domain_id = {domain_id},'
                              f' request_domain_id = {request_domain_id})')
                raise ERROR_PERMISSION_DENIED()

            if request_domain_id and request_domain_id != domain_id:
                _LOGGER.debug(f'[_check_domain_scope] domain_id is not allowed.'
                              f' (user_id = {user_id}, user_domain_id = {domain_id},'
                              f' request_domain_id = {request_domain_id})')
                raise ERROR_PERMISSION_DENIED()

    @staticmethod
    def _check_user_scope(user_id, domain_id, role_type, request_user_id, require_user_id):
        if role_type in ['PROJECT', 'USER']:
            if require_user_id and request_user_id is None:
                _LOGGER.debug(f'[_check_user_scope] user_id is required.'
                              f' (user_id = {user_id}, user_domain_id = {domain_id},'
                              f' request_user_id = {request_user_id})')
                raise ERROR_PERMISSION_DENIED()

            if request_user_id and request_user_id != user_id:
                _LOGGER.debug(f'[_check_user_scope] user role_type can only access self resource.'
                              f' (user_id = {user_id}, user_domain_id = {domain_id},'
                              f' request_user_id = {request_user_id})')
                raise ERROR_PERMISSION_DENIED()

    @staticmethod
    def _check_project_scope(user_id, domain_id, role_type, projects, project_groups,
                             request_project_id, request_project_group_id,
                             require_project_id, require_project_group_id):
        if role_type == 'PROJECT':
            if require_project_id and request_project_id is None:
                _LOGGER.debug(f'[_check_project_scope] project_id is required.'
                              f' (user_id = {user_id}, user_domain_id = {domain_id},'
                              f' request_project_id = {request_project_id})')
                raise ERROR_PERMISSION_DENIED()

            if require_project_group_id and request_project_group_id is None:
                _LOGGER.debug(f'[_check_project_scope] project_group_id is required.'
                              f' (user_id = {user_id}, user_domain_id = {domain_id},'
                              f' request_project_group_id = {request_project_group_id})')
                raise ERROR_PERMISSION_DENIED()

            if request_project_id and request_project_id not in projects:
                _LOGGER.debug(f'[_check_project_scope] project_id is not allowed.'
                              f' (user_id = {user_id}, user_domain_id = {domain_id},'
                              f' request_project_id = {request_project_id})')
                raise ERROR_PERMISSION_DENIED()

            if request_project_group_id and request_project_group_id not in project_groups:
                _LOGGER.debug(f'[_check_project_scope] project_group_id is not allowed.'
                              f' (user_id = {user_id}, user_domain_id = {domain_id},'
                              f' request_project_group_id = {request_project_group_id})')
                raise ERROR_PERMISSION_DENIED()

    @staticmethod
    def _match_permission(service, api_class, method, permission):
        if fnmatch.fnmatch(f'{service}.{api_class}.{method}', permission):
            return True
        else:
            return False



from spaceone.api.identity.v1 import api_key_pb2, api_key_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class APIKey(BaseAPI, api_key_pb2_grpc.APIKeyServicer):

    pb2 = api_key_pb2
    pb2_grpc = api_key_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('APIKeyService', metadata) as api_key_svc:
            api_key_vo, api_key = api_key_svc.create(params)
            return self.locator.get_info('APIKeyInfo', api_key_vo, api_key=api_key)

    def enable(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('APIKeyService', metadata) as api_key_svc:
            data = api_key_svc.enable(params)
            return self.locator.get_info('APIKeyInfo', data)

    def disable(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('APIKeyService', metadata) as api_key_svc:
            data = api_key_svc.disable(params)
            return self.locator.get_info('APIKeyInfo', data)

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('APIKeyService', metadata) as api_key_svc:
            api_key_svc.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('APIKeyService', metadata) as api_key_svc:
            data = api_key_svc.get(params)
            return self.locator.get_info('APIKeyInfo', data)

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('APIKeyService', metadata) as api_key_svc:
            api_key_vos, total_count = api_key_svc.list(params)
            return self.locator.get_info('APIKeysInfo', api_key_vos, total_count, minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('APIKeyService', metadata) as api_key_svc:
            return self.locator.get_info('StatisticsInfo', api_key_svc.stat(params))



from spaceone.core.service import *
from spaceone.identity.manager.endpoint_manager import EndpointManager


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class EndpointService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.endpoint_mgr: EndpointManager = self.locator.get_manager('EndpointManager')

    @transaction(append_meta={'authorization.scope': 'PUBLIC'})
    @append_query_filter(['service'])
    @append_keyword_filter(['service'])
    def list(self, params):
        """
        Args:
            params (dict): {
                    'query': 'dict (spaceone.api.core.v1.Query)',
                    'service': 'str',
                    'endpoint_type': 'str'
                }

        Returns:
            results (list): list of endpoint_vo
            total_count (int)
        """
        return self.endpoint_mgr.list_endpoints(params.get('query', {}), params.get('endpoint_type', 'public'))



from mongoengine import *
from spaceone.core.error import *
from spaceone.core.model.mongo_model import MongoModel


class Policy(MongoModel):
    policy_id = StringField(max_length=40, generate_id='policy', unique_with=['policy_type', 'domain_id'])
    name = StringField(max_length=255)
    policy_type = StringField(max_length=20, default='CUSTOM', choices=('CUSTOM', 'MANAGED'))
    permissions = ListField(StringField())
    tags = DictField()
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(default=None, null=True)

    meta = {
        'updatable_fields': [
            'name',
            'permissions',
            'tags',
            'updated_at'
        ],
        'minimal_fields': [
            'policy_id',
            'name'
        ],
        'ordering': ['name'],
        'indexes': [
            'policy_id',
            'policy_type',
            'domain_id',
        ]
    }



import logging

from spaceone.core.cache import cacheable
from spaceone.core.manager import BaseManager
from spaceone.identity.lib.key_generator import KeyGenerator
from spaceone.identity.model.api_key_model import APIKey
from spaceone.identity.model.domain_secret_model import DomainSecret

_LOGGER = logging.getLogger(__name__)


class APIKeyManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_key_model: APIKey = self.locator.get_model('APIKey')
        self.secret_model: DomainSecret = self.locator.get_model('DomainSecret')

    def create_api_key(self, user_vo, domain_id):
        def _rollback(api_key_vo):
            _LOGGER.info(f'[create_api_key._rollback] Delete api_key : {api_key_vo.api_key_id}')
            api_key_vo.delete()

        params = {
            'user_id': user_vo.user_id,
            'user': user_vo,
            'domain_id': domain_id
        }
        api_key_vo: APIKey = self.api_key_model.create(params)
        self.transaction.add_rollback(_rollback, api_key_vo)

        prv_jwk = self._query_domain_secret(domain_id)

        key_gen = KeyGenerator(prv_jwk=prv_jwk,
                               domain_id=domain_id,
                               audience=user_vo.user_id)

        api_key = key_gen.generate_api_key(api_key_vo.api_key_id, user_vo.user_type)
        return api_key_vo, api_key

    def delete_api_key(self, api_key_id, domain_id):
        api_key_vo = self.get_api_key(api_key_id, domain_id)
        api_key_vo.delete()

    def enable_api_key(self, api_key_id, domain_id):
        def _rollback(old_data):
            _LOGGER.info(f'[enable_api_key._rollback] Revert Data: {old_data}')
            api_key_vo.update(old_data)

        api_key_vo: APIKey = self.get_api_key(api_key_id, domain_id)
        if api_key_vo.state != 'ENABLED':
            self.transaction.add_rollback(_rollback, api_key_vo.to_dict())
            api_key_vo.update({'state': 'ENABLED'})

        return api_key_vo

    def disable_api_key(self, api_key_id, domain_id):
        def _rollback(old_data):
            _LOGGER.info(f'[disable_api_key._rollback] Revert Data: {old_data}')
            api_key_vo.update(old_data)

        api_key_vo: APIKey = self.get_api_key(api_key_id, domain_id)
        if api_key_vo.state != 'DISABLED':
            self.transaction.add_rollback(_rollback, api_key_vo.to_dict())
            api_key_vo.update({'state': 'DISABLED'})

        return api_key_vo

    def get_api_key(self, api_key_id, domain_id, only=None):
        return self.api_key_model.get(api_key_id=api_key_id, domain_id=domain_id, only=only)

    def list_api_keys(self, query):
        return self.api_key_model.query(**query)

    def stat_api_keys(self, query):
        return self.api_key_model.stat(**query)

    @cacheable(key='api-key:{domain_id}', expire=60)
    def _query_domain_secret(self, domain_id: str) -> DomainSecret:
        domain_secret = self.secret_model.get(domain_id=domain_id)
        return domain_secret.prv_jwk



from mongoengine import *
from spaceone.core.model.mongo_model import MongoModel
from spaceone.identity.model.project_model import Project


class ServiceAccount(MongoModel):
    service_account_id = StringField(max_length=40, generate_id='sa', unique=True)
    name = StringField(max_length=255, unique_with='domain_id')
    data = DictField()
    service_account_type = StringField(max_length=40, choices=('TRUSTED', 'GENERAL'), default='GENERAL')
    provider = StringField(max_length=40)
    project = ReferenceField('Project', null=True, default=None, reverse_delete_rule=DENY)
    project_id = StringField(max_length=40, default=None, null=True)
    trusted_service_account_id = StringField(max_length=40, null=True, default=None)
    tags = DictField()
    scope = StringField(max_length=40, choices=('DOMAIN', 'PROJECT'), default='PROJECT')
    domain_id = StringField(max_length=255)
    created_at = DateTimeField(auto_now_add=True)

    meta = {
        'updatable_fields': [
            'name',
            'data',
            'project',
            'project_id',
            'trusted_service_account_id',
            'tags'
        ],
        'minimal_fields': [
            'service_account_id',
            'service_account_type',
            'name',
            'provider'
        ],
        'change_query_keys': {
            'user_projects': 'project_id',
            'project_id': 'project_id'
        },
        'reference_query_keys': {
            'project': Project
        },
        'ordering': ['name'],
        'indexes': [
            # 'service_account_id',
            'name',
            'provider',
            'project',
            'project_id',
            'trusted_service_account_id',
            'scope',
            'domain_id'
        ]
    }



import functools
from spaceone.api.identity.v1 import role_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.identity.model.role_model import Role, RolePolicy, PagePermission

__all__ = ['RoleInfo', 'RolesInfo']


def RolePolicyInfo(role_policy_vo: RolePolicy):
    role_policy_info = {
        'policy_type': role_policy_vo.policy_type,
        'policy_id': role_policy_vo.policy_id
    }

    return role_policy_info


def PagePermissionInfo(page_perm_vo: PagePermission):
    page_perm_info = {
        'page': page_perm_vo.page,
        'permission': page_perm_vo.permission
    }

    return page_perm_info


def RoleInfo(role_vo: Role, minimal=False):
    # Temporary Code for Migration
    policies = []
    is_changed = False
    for policy_vo in role_vo.policies:
        if not policy_vo.policy_id:
            policy_vo.policy_id = policy_vo.policy.policy_id
            is_changed = True

        policies.append({
            'policy_type': policy_vo.policy_type,
            'policy': policy_vo.policy,
            'policy_id': policy_vo.policy_id
        })

    if is_changed:
        role_vo = role_vo.update({'policies': policies})

    info = {
        'role_id': role_vo.role_id,
        'name': role_vo.name,
        'role_type': role_vo.role_type,
        'page_permissions': list(map(lambda page_perm_vo: PagePermissionInfo(page_perm_vo), role_vo.page_permissions))
    }

    if not minimal:
        info.update({
            'policies': list(map(lambda policy: RolePolicyInfo(policy), role_vo.policies)),
            'tags': change_struct_type(role_vo.tags),
            'domain_id': role_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(role_vo.created_at)
        })

    return role_pb2.RoleInfo(**info)


def RolesInfo(role_vos, total_count, **kwargs):
    results = list(map(functools.partial(RoleInfo, **kwargs), role_vos))

    return role_pb2.RolesInfo(results=results, total_count=total_count)



from spaceone.core.error import *


class ERROR_EXIST_CHILD_PROJECT_GROUP(ERROR_BASE):
    _message = 'Child project group is exist. project_group_id = {project_group_id}'


class ERROR_ALREADY_EXIST_USER_IN_PROJECT_GROUP(ERROR_BASE):
    _message = 'A user "{user_id}" is already exist in project group({project_group_id}).'


class ERROR_ALREADY_EXIST_USER_IN_PROJECT(ERROR_BASE):
    _message = 'A user "{user_id}" is already exist in project({project_id}).'


class ERROR_NOT_FOUND_USER_IN_PROJECT_GROUP(ERROR_BASE):
    _message = 'A user "{user_id}" is not exist in project group({project_group_id}).'


class ERROR_NOT_FOUND_USER_IN_PROJECT(ERROR_BASE):
    _message = 'A user "{user_id}" is not exist in project({project_id}).'


class ERROR_ONLY_PROJECT_ROLE_TYPE_ALLOWED(ERROR_INVALID_ARGUMENT):
    _message = 'Only project role type can be allowed.'



from spaceone.identity.service.authorization_service import AuthorizationService
from spaceone.identity.service.user_service import UserService
from spaceone.identity.service.role_service import RoleService
from spaceone.identity.service.role_binding_service import RoleBindingService
from spaceone.identity.service.api_key_service import APIKeyService
from spaceone.identity.service.domain_service import DomainService
from spaceone.identity.service.domain_owner_service import DomainOwnerService
from spaceone.identity.service.provider_service import ProviderService
from spaceone.identity.service.service_account_service import ServiceAccountService
from spaceone.identity.service.policy_service import PolicyService
from spaceone.identity.service.token_service import TokenService
from spaceone.identity.service.project_group_service import ProjectGroupService
from spaceone.identity.service.project_service import ProjectService
from spaceone.identity.service.endpoint_service import EndpointService



from spaceone.api.identity.v1 import token_pb2, token_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class Token(BaseAPI, token_pb2_grpc.TokenServicer):

    pb2 = token_pb2
    pb2_grpc = token_pb2_grpc

    def issue(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('TokenService', metadata) as token_svc:
            data = token_svc.issue(params)
            return self.locator.get_info('TokenInfo', data)

    def refresh(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('TokenService', metadata) as token_svc:
            data = token_svc.refresh(params)
            return self.locator.get_info('TokenInfo', data)



from spaceone.api.identity.v1 import endpoint_pb2, endpoint_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class Endpoint(BaseAPI, endpoint_pb2_grpc.EndpointServicer):

    pb2 = endpoint_pb2
    pb2_grpc = endpoint_pb2_grpc

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('EndpointService', metadata) as endpoint_svc:
            endpoint_vos, total_count = endpoint_svc.list(params)
            return self.locator.get_info('EndpointsInfo', endpoint_vos,
                                         total_count, minimal=self.get_minimal(params))



import logging

from spaceone.core.error import *
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.identity.model.service_account_model import ServiceAccount

_LOGGER = logging.getLogger(__name__)


class ServiceAccountManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.service_account_model: ServiceAccount = self.locator.get_model('ServiceAccount')

    def create_service_account(self, params):
        def _rollback(service_account_vo):
            _LOGGER.info(f'[create_service_account._rollback] '
                         f'Create service_account : {service_account_vo.name} '
                         f'({service_account_vo.service_account_id})')
            service_account_vo.delete()

        service_account_vo: ServiceAccount = self.service_account_model.create(params)
        self.transaction.add_rollback(_rollback, service_account_vo)

        return service_account_vo

    def update_service_account(self, params):
        service_account_vo: ServiceAccount = self.get_service_account(params['service_account_id'],
                                                                      params['domain_id'])

        return self.update_service_account_by_vo(params, service_account_vo)

    def update_service_account_by_vo(self, params, service_account_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_service_account._rollback] Revert Data : '
                         f'{old_data["service_account_id"]}')
            service_account_vo.update(old_data)

        self.transaction.add_rollback(_rollback, service_account_vo.to_dict())

        return service_account_vo.update(params)

    def delete_service_account(self, service_account_id, domain_id):
        service_account_vo: ServiceAccount = self.get_service_account(service_account_id, domain_id)
        service_account_vo.delete()

    def get_service_account(self, service_account_id, domain_id, only=None):
        return self.service_account_model.get(service_account_id=service_account_id, domain_id=domain_id, only=only)

    def list_service_accounts(self, query={}):
        return self.service_account_model.query(**query)

    def stat_service_accounts(self, query):
        return self.service_account_model.stat(**query)

    def update_secret_project(self, service_account_id, project_id, domain_id):
        secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='secret')
        response = self._list_secrets(secret_connector, service_account_id, domain_id)
        secrets = response.get('results', [])

        for secret_info in secrets:
            secret_connector.dispatch('Secret.update', {
                'secret_id': secret_info['secret_id'],
                'project_id': project_id,
                'domain_id': domain_id
            })

    def release_secret_project(self, service_account_id, domain_id):
        secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='secret')
        response = self._list_secrets(secret_connector, service_account_id, domain_id)
        secrets = response.get('results', [])

        for secret_info in secrets:
            secret_connector.dispatch('Secret.update', {
                'secret_id': secret_info['secret_id'],
                'release_project': True,
                'domain_id': domain_id
            })

    def delete_service_account_secrets(self, service_account_id, domain_id, service_account_type):
        secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='secret')
        if service_account_type == 'TRUSTED':
            response = self._list_trusted_secrets(secret_connector, service_account_id, domain_id)
            for trusted_secret_info in response.get('results', []):
                secret_connector.dispatch('TrustedSecret.delete', {
                    'trusted_secret_id': trusted_secret_info['trusted_secret_id'],
                    'domain_id': domain_id
                })
        else:
            response = self._list_secrets(secret_connector, service_account_id, domain_id)
            for secret_info in response.get('results', []):
                secret_connector.dispatch('Secret.delete', {
                    'secret_id': secret_info['secret_id'],
                    'domain_id': domain_id
                })

    def get_all_service_account_ids_using_secret(self, domain_id):
        secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='secret')
        response = secret_connector.dispatch('Secret.stat', {
            'query': {
                'distinct': 'service_account_id',
                'filter': [
                    {
                        'k': 'service_account_id',
                        'v': None,
                        'o': 'not'
                    }
                ]
            },
            'domain_id': domain_id
        })

        return response.get('results', [])

    def check_service_account_secrets(self, service_account_id, domain_id, service_account_type):
        secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='secret')

        if service_account_type == 'TRUSTED':
            response = self._list_trusted_secrets(secret_connector, service_account_id, domain_id)
            total_count = response.get('total_count', 0)

            if total_count > 0:
                raise ERROR_EXIST_RESOURCE(parent='ServiceAccount', child='TrustedSecret')
        else:
            response = self._list_secrets(secret_connector, service_account_id, domain_id)

            total_count = response.get('total_count', 0)

            if total_count > 0:
                raise ERROR_EXIST_RESOURCE(parent='ServiceAccount', child='Secret')

    @staticmethod
    def _list_secrets(secret_connector, service_account_id, domain_id):
        return secret_connector.dispatch('Secret.list', {
            'service_account_id': service_account_id,
            'domain_id': domain_id
        })

    @staticmethod
    def _list_trusted_secrets(secret_connector, service_account_id, domain_id):
        return secret_connector.dispatch('TrustedSecret.list', {
            'service_account_id': service_account_id,
            'domain_id': domain_id
        })


import logging
from datetime import datetime
from spaceone.identity.connector import AuthPluginConnector
from spaceone.identity.error.error_authentication import *
from spaceone.identity.error.error_user import ERROR_USER_STATUS_CHECK_FAILURE
from spaceone.identity.manager.user_manager import UserManager
from spaceone.identity.manager.token_manager import JWTManager
from spaceone.identity.manager.domain_manager import DomainManager
from spaceone.identity.model import Domain, User

_LOGGER = logging.getLogger(__name__)


class ExternalTokenManager(JWTManager):
    domain: Domain = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.domain_mgr: DomainManager = self.locator.get_manager('DomainManager')
        self.user_mgr: UserManager = self.locator.get_manager('UserManager')

    def authenticate(self, user_id, domain_id, credentials):
        _LOGGER.debug(f'[authenticate] domain_id: {domain_id}')

        # Add User ID for External Authentication
        if user_id:
            credentials[user_id] = user_id

        self.domain: Domain = self.domain_mgr.get_domain(domain_id)

        self._check_domain_state()

        endpoint = self.domain_mgr.get_auth_plugin_endpoint_by_vo(self.domain)
        auth_user_info = self._authenticate_with_plugin(endpoint, credentials)

        _LOGGER.info(f'[authenticate] Authentication success. (user_id={auth_user_info.get("user_id")})')

        auto_user_sync = self.domain.plugin_info.options.get('auto_user_sync', False)

        self._verify_user_from_plugin_user_info(auth_user_info, domain_id, auto_user_sync)
        self._check_user_state()

        self.is_authenticated = True

    def issue_token(self, **kwargs):
        if self.is_authenticated is False:
            raise ERROR_NOT_AUTHENTICATED()

        if self.user.state == 'PENDING':
            self.user: User = self.user.update({'state': 'ENABLED'})

        # Issue token
        access_token = self.issue_access_token('USER', self.user.user_id, self.user.domain_id, **kwargs)
        refresh_token = self.issue_refresh_token('USER', self.user.user_id, self.user.domain_id, **kwargs)

        # Update user's last_accessed_at field
        self.user.update({'last_accessed_at': datetime.utcnow()})

        return {
            'access_token': access_token,
            'refresh_token': refresh_token
        }

    def refresh_token(self, user_id, domain_id, **kwargs):
        self.user: User = self.user_mgr.get_user(user_id, domain_id)
        self._check_user_state()

        return self.issue_token(**kwargs)

    def _verify_user_from_plugin_user_info(self, auth_user_info, domain_id, auto_user_sync=False):
        if 'user_id' not in auth_user_info:
            _LOGGER.error(f'[_verify_user_from_plugin_user_info] does not return user_id from plugin user info.')
            raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(message='plugin response is invalid.')

        user_id = auth_user_info['user_id']
        state = auth_user_info.get('state', 'ENABLED')

        user_vos = self.user_mgr.filter_users(user_id=user_id, domain_id=domain_id)

        if user_vos.count() > 0:
            self.user: User = user_vos[0]
        else:
            if auto_user_sync:
                name = auth_user_info.get('name')
                email = auth_user_info.get('email')
                self.user: User = self._create_external_user(user_id, state, domain_id, name, email)
            else:
                raise ERROR_NOT_FOUND(key='user_id', value=user_id)

    def _authenticate_with_plugin(self, endpoint, credentials):
        options = self.domain.plugin_info.options

        auth_plugin_conn: AuthPluginConnector = self.locator.get_connector('AuthPluginConnector')
        auth_plugin_conn.initialize(endpoint)

        return auth_plugin_conn.call_login(credentials, options, {})

    def _check_domain_state(self):
        if not self.domain.plugin_info:
            _LOGGER.error('[_get_token_manager] This domain does not allow external authentication.')
            raise ERROR_AUTHENTICATION_FAILURE(user_id=self.user.user_id)

    def _check_user_state(self):
        if self.user.state not in ['ENABLED', 'PENDING']:
            raise ERROR_USER_STATUS_CHECK_FAILURE(user_id=self.user.user_id)

        if self.user.backend != 'EXTERNAL':
            raise ERROR_NOT_FOUND(key='user_id', value=self.user.user_id)

    def _create_external_user(self, user_id, state, domain_id, name=None, email=None):
        _LOGGER.error(f'[_create_external_user] create user on first login: {user_id}')
        return self.user_mgr.create_user({
            'user_id': user_id,
            'state': state,
            'name': name,
            'email': email,
            'user_type': 'USER',
            'backend': 'EXTERNAL',
            'domain_id': domain_id
        }, self.domain, is_first_login_user=True)



import functools
from spaceone.api.identity.v1 import policy_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.identity.model.policy_model import Policy

__all__ = ['PolicyInfo', 'PoliciesInfo']


def PolicyInfo(policy_vo: Policy, minimal=False):
    info = {
        'policy_id': policy_vo.policy_id,
        'name': policy_vo.name
    }

    if not minimal:
        info.update({
            'permissions': change_list_value_type(policy_vo.permissions),
            'tags': change_struct_type(policy_vo.tags),
            'domain_id': policy_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(policy_vo.created_at)
        })

    return policy_pb2.PolicyInfo(**info)


def PoliciesInfo(policy_vos, total_count, **kwargs):
    results = list(map(functools.partial(PolicyInfo, **kwargs), policy_vos))

    return policy_pb2.PoliciesInfo(results=results, total_count=total_count)



ROOT_DOMAIN_NAME = 'root'
EMAIL_CONSOLE_DOMAIN = ''
EMAIL_SERVICE_NAME = 'Cloudforet'
# ACCESS_TOKEN (default) / PASSWORD
RESET_PASSWORD_TYPE = 'ACCESS_TOKEN'
DATABASE_AUTO_CREATE_INDEX = True
DATABASES = {
    'default': {
        # 'db': '',
        # 'host': '',
        # 'port': 0,
        # 'username': '',
        # 'password': '',
        # 'ssl': False,
        # 'ssl_ca_certs': ''
    }
}

CACHES = {
    'default': {},
    'local': {
        'backend': 'spaceone.core.cache.local_cache.LocalCache',
        'max_size': 128,
        'ttl': 300
    }
}

IDENTITY = {
    'token': {
        'verify_code_timeout': 3600,
        'temporary_token_timeout': 86400,  # 24 hours
        'token_timeout': 1200,
        'refresh_timeout': 1800,
        'refresh_ttl': 18,
        'refresh_once': False
    }
}

HANDLERS = {
    # 'authentication': [{
    #     'backend': 'spaceone.core.handler.authentication_handler.AuthenticationGRPCHandler',
    #     'uri': 'grpc://localhost:50051/v1/Domain/get_public_key'
    # }],
    # 'authorization': [{
    #     'backend': 'spaceone.core.handler.authorization_handler.AuthorizationGRPCHandler',
    #     'uri': 'grpc://localhost:50051/v1/Authorization/verify'
    # }],
    # 'mutation': [{
    #     'backend': 'spaceone.core.handler.mutation_handler.SpaceONEMutationHandler'
    # }],
    # 'event': []
}

CONNECTORS = {
    'SpaceConnector': {
        'backend': 'spaceone.core.connector.space_connector.SpaceConnector',
        'endpoints': {
            'plugin': 'grpc://plugin:50051',
            'secret': 'grpc://secret:50051',
            'repository': 'grpc://repository:50051'
        }
    },
    'SMTPConnector': {
        'host': 'smtp.mail.com',
        'port': '1234',
        'user': 'cloudforet',
        'password': '1234',
        'from_email': 'support@cloudforet.com'
    }
}

ENDPOINTS = [
    # {
    #     'service': 'identity',
    #     'name': 'Identity Service',
    #     'endpoint': 'grpc://<endpoint>>:<port>/v1'
    # },
]

# Internal Endpoint
INTERNAL_ENDPOINTS = [
    {
        'service': 'identity',
        'name': 'Identity Service',
        'endpoint': 'grpc://identity.spaceone.svc.cluster.local:50051/v1'
    },
    {
        'service': 'secret',
        'name': 'Secret Service',
        'endpoint': 'grpc://secret.spaceone.svc.cluster.local:50051/v1'
    },
    {
        'service': 'repository',
        'name': 'Repository Service',
        'endpoint': 'grpc://repository.spaceone.svc.cluster.local:50051/v1'
    },
    {
        'service': 'plugin',
        'name': 'Plugin Service',
        'endpoint': 'grpc://plugin.spaceone.svc.cluster.local:50051/v1'
    },
    {
        'service': 'config',
        'name': 'Config Service',
        'endpoint': 'grpc://config.spaceone.svc.cluster.local:50051/v1'
    },
    {
        'service': 'inventory',
        'name': 'Inventory Service',
        'endpoint': 'grpc://inventory.spaceone.svc.cluster.local:50051/v1'
    },
    {
        'service': 'monitoring',
        'name': 'Monitoring Service',
        'endpoint': 'grpc://monitoring.spaceone.svc.cluster.local:50051/v1'
    },
    {
        'service': 'statistics',
        'name': 'Statistics Service',
        'endpoint': 'grpc://statistics.spaceone.svc.cluster.local:50051/v1'
    }
]



import functools
from spaceone.api.identity.v1 import user_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.identity.model.user_model import User

__all__ = ['UserInfo', 'UsersInfo']


def UserInfo(user_vo: User, minimal=False):
    info = {
        'user_id': user_vo.user_id,
        'name': user_vo.name,
        'state': user_vo.state,
        'user_type': user_vo.user_type
    }

    if not minimal:
        info.update({
            'email': user_vo.email,
            'email_verified': user_vo.email_verified,
            'backend': user_vo.backend,
            'required_actions': user_vo.required_actions,
            'language': user_vo.language,
            'timezone': user_vo.timezone,
            'tags': change_struct_type(user_vo.tags),
            'domain_id': user_vo.domain_id,
            'last_accessed_at': utils.datetime_to_iso8601(user_vo.last_accessed_at),
            'created_at': utils.datetime_to_iso8601(user_vo.created_at)
        })

    return user_pb2.UserInfo(**info)


def UsersInfo(user_vos, total_count, **kwargs):
    results = list(map(functools.partial(UserInfo, **kwargs), user_vos))
    return user_pb2.UsersInfo(results=results, total_count=total_count)



import logging
from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.identity.model.project_group_model import ProjectGroup

_LOGGER = logging.getLogger(__name__)


class ProjectGroupManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.project_group_model: ProjectGroup = self.locator.get_model('ProjectGroup')

    def create_project_group(self, params):
        def _rollback(project_group_vo):
            _LOGGER.info(
                f'[create_project_group._rollback] Delete project group : {project_group_vo.name} ({project_group_vo.project_group_id})')
            project_group_vo.delete()

        project_group_vo: ProjectGroup = self.project_group_model.create(params)
        self.transaction.add_rollback(_rollback, project_group_vo)

        if params.get('parent_project_group') is not None:
            parent_project_group_id = params['parent_project_group_id']
            cache.delete_pattern(f'project-path:*{parent_project_group_id}*')
            cache.delete_pattern(f'role-bindings:*{parent_project_group_id}*')
            cache.delete_pattern(f'user-scopes:*{parent_project_group_id}*')
            self._delete_parent_project_group_cache(params['parent_project_group'])

        return project_group_vo

    def update_project_group(self, params):
        project_group_vo = self.get_project_group(params['project_group_id'], params['domain_id'])
        return self.update_project_group_by_vo(params, project_group_vo)

    def update_project_group_by_vo(self, params, project_group_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_project_group._rollback] Revert Data : {old_data["name"]} ({old_data["project_group_id"]})')
            project_group_vo.update(old_data)

        self.transaction.add_rollback(_rollback, project_group_vo.to_dict())

        if params.get('parent_project_group') is not None:
            parent_project_group_id = params['parent_project_group_id']
            project_group_id = project_group_vo.project_group_id

            if parent_project_group_id is not None:
                cache.delete_pattern(f'project-path:*{parent_project_group_id}*')
                cache.delete_pattern(f'role-bindings:*{parent_project_group_id}*')
                cache.delete_pattern(f'user-scopes:*{parent_project_group_id}*')
                self._delete_parent_project_group_cache(params['parent_project_group'])

            cache.delete_pattern(f'project-path:*{project_group_id}*')
            cache.delete_pattern(f'role-bindings:*{project_group_id}*')
            cache.delete_pattern(f'user-scopes:*{project_group_id}*')
            self._delete_parent_project_group_cache(project_group_vo)

        return project_group_vo.update(params)

    def delete_project_group(self, project_group_id, domain_id):
        project_group_vo = self.get_project_group(project_group_id, domain_id)
        self.delete_project_group_by_vo(project_group_vo)

    def delete_project_group_by_vo(self, project_group_vo):
        project_group_vo.delete()

        cache.delete_pattern(f'project-path:*{project_group_vo.project_group_id}*')
        cache.delete_pattern(f'role-bindings:*{project_group_vo.project_group_id}*')
        cache.delete_pattern(f'role-bindings:*:')
        cache.delete_pattern(f'user-scopes:*{project_group_vo.project_group_id}*')
        self._delete_parent_project_group_cache(project_group_vo)

    def get_project_group(self, project_group_id, domain_id, only=None):
        return self.project_group_model.get(project_group_id=project_group_id, domain_id=domain_id, only=only)

    def filter_project_groups(self, **conditions):
        return self.project_group_model.filter(**conditions)

    def list_project_groups(self, query):
        return self.project_group_model.query(**query)

    def stat_project_groups(self, query):
        return self.project_group_model.stat(**query)

    def _delete_parent_project_group_cache(self, project_group_vo):
        cache.delete_pattern(f'project-group-children:*{project_group_vo.project_group_id}')
        if project_group_vo.parent_project_group:
            self._delete_parent_project_group_cache(project_group_vo.parent_project_group)



import logging

from spaceone.core.service import *
from spaceone.core import utils
from spaceone.identity.error import *
from spaceone.identity.manager import DomainManager
from spaceone.identity.manager.domain_secret_manager import DomainSecretManager
from spaceone.identity.model import Domain

_LOGGER = logging.getLogger(__name__)


@authentication_handler(exclude=['create', 'list', 'get_public_key'])
@authorization_handler(exclude=['create', 'list', 'get_public_key'])
@mutation_handler(exclude=['create', 'list', 'get_public_key'])
@event_handler
class DomainService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.domain_mgr: DomainManager = self.locator.get_manager('DomainManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['name'])
    def create(self, params):
        """ Create domain

        Args:
            params (dict): {
                'name': 'str',
                'config': 'dict',
                'plugin_info': 'dict',
                'tags': 'dict'
            }

        Returns:
            domain_vo (object)
        """

        # Create Domain
        domain_vo: Domain = self.domain_mgr.create_domain(params)

        # Create domain secret
        domain_secret_mgr: DomainSecretManager = self._get_domain_secret_manager()
        domain_secret_mgr.create_domain_secret(domain_vo.domain_id)

        return domain_vo

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    def update(self, params):
        """ Update domain

        Args:
            params (dict): {
                'domain_id': 'str',
                'config': 'dict',
                'tags': 'dict'
            }

        Returns:
            domain_vo (object)
        """

        return self.domain_mgr.update_domain(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    def change_auth_plugin(self, params):
        """ Change domain auth plugin

        Args:
            params (dict): {
                'domain_id': 'str',
                'plugin_info': 'dict',
                'release_auth_plugin': 'bool'
            }

        Returns:
            domain_vo (object)
        """

        domain_id = params['domain_id']
        plugin_info = params.get('plugin_info')
        release_auth_plugin = params.get('release_auth_plugin', False)

        if release_auth_plugin:
            # release auth plugin
            _LOGGER.debug(f'[change_auth_plugin] release auth plugin')
            return self.domain_mgr.release_auth_plugin(domain_id)
        else:
            if plugin_info is None:
                raise ERROR_REQUIRED_PARAMETER(key='plugin_info')

            _LOGGER.debug(f'[change_auth_plugin] update plugin_info: {plugin_info}')
            return self.domain_mgr.change_auth_plugin(domain_id, plugin_info)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    def update_plugin(self, params):
        """ Update Plugin
        Args:
            params (dict): {
                'domain_id': 'str',
                'version': 'str',
                'options': 'dict',
                'upgrade_mode': 'str',
            }

        Returns:
            domain_vo (object)
        """

        domain_id = params['domain_id']
        version = params.get('version')
        options = params.get('options')
        upgrade_mode = params.get('upgrade_mode')

        return self.domain_mgr.update_domain_plugin(domain_id, version, options, upgrade_mode)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    def verify_plugin(self, params):
        """ Update Plugin
        Args:
            params (dict): {
                'domain_id': 'str',
            }

        Returns:
            domain_vo (object)
        """
        domain_id = params['domain_id']
        return self.domain_mgr.verify_auth_plugin(domain_id)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    def delete(self, params):
        """ Delete domain

        Args:
            params (dict): {
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.domain_mgr.delete_domain(params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    def enable(self, params):
        """ Enable domain

        Args:
            params (dict): {
                'domain_id': 'str'
            }

        Returns:
            domain_vo (object)
        """

        return self.domain_mgr.enable_domain(params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    def disable(self, params):
        """ Disable domain

        Args:
            params (dict): {
                'domain_id': 'str'
            }

        Returns:
            domain_vo (object)
        """

        return self.domain_mgr.disable_domain(params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    def get(self, params):
        """ Disable domain

        Args:
            params (dict): {
                'domain_id': 'str',
                'only': 'list'
            }

        Returns:
            domain_vo (object)
        """

        return self.domain_mgr.get_domain(params['domain_id'], params.get('only'))

    @transaction(append_meta={'auth.scope': 'SYSTEM'})
    @check_required(['domain_id'])
    def get_public_key(self, params):
        """ Get domain's public key for authentication

        Args:
            params (dict): {
                'domain_id': 'str'
            }

        Returns:
            result (dict): {
                'pub_jwk': 'str',
                'domain_id': 'str'
            }
        """

        domain_id = params['domain_id']
        domain_secret_mgr: DomainSecretManager = self._get_domain_secret_manager()
        pub_jwk = domain_secret_mgr.get_domain_public_key(domain_id=domain_id)

        return {
            'pub_jwk': pub_jwk,
            'domain_id': domain_id
        }

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @append_query_filter(['domain_id', 'name'])
    @append_keyword_filter(['domain_id', 'name'])
    def list(self, params):
        """ List api keys

        Args:
            params (dict): {
                'domain_id': 'str',
                'name': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of domain_vo'
            total_count (int)
        """

        query = params.get('query', {})
        return self.domain_mgr.list_domains(query)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query'])
    @append_keyword_filter(['domain_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list): 'list of statistics data'
            total_count (int)
        """

        query = params.get('query', {})
        return self.domain_mgr.stat_domains(query)

    def _get_domain_secret_manager(self):
        return self.locator.get_manager('DomainSecretManager')



import logging
from spaceone.core.service import *
from spaceone.core import utils
from spaceone.identity.error.error_project import *
from spaceone.identity.manager.domain_manager import DomainManager
from spaceone.identity.manager.user_manager import UserManager
from spaceone.identity.model.project_group_model import ProjectGroup
from spaceone.identity.manager.project_manager import ProjectManager
from spaceone.identity.manager.project_group_manager import ProjectGroupManager
from spaceone.identity.manager.role_manager import RoleManager
from spaceone.identity.manager.role_binding_manager import RoleBindingManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class ProjectGroupService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.project_group_mgr: ProjectGroupManager = self.locator.get_manager('ProjectGroupManager')

    @transaction(append_meta={
        'authorization.scope': 'PROJECT',
        'authorization.project_group_id': 'parent_project_group_id',
        'authorization.require_project_group_id': True
    })
    @check_required(['name', 'domain_id'])
    def create(self, params):
        """ Create project group

        Args:
            params (dict): {
                'name': 'str',
                'parent_project_group_id': 'str',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            project_group_vo (object)
        """

        params['created_by'] = self.transaction.get_meta('user_id')

        if 'parent_project_group_id' in params:
            params['parent_project_group'] = self._get_parent_project_group(params['parent_project_group_id'],
                                                                            params['domain_id'])
        else:
            params['parent_project_group'] = None

        return self.project_group_mgr.create_project_group(params)

    @transaction(append_meta={
        'authorization.scope': 'PROJECT',
        'authorization.project_group_id': 'parent_project_group_id'
    })
    @check_required(['project_group_id', 'domain_id'])
    def update(self, params):
        """ Update project group

        Args:
            params (dict): {
                'project_group_id': 'str',
                'name': 'str',
                'parent_project_group_id': 'str',
                'release_parent_project_group': 'bool',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            project_group_vo (object)
        """

        domain_id = params['domain_id']
        release_parent_project_group = params.get('release_parent_project_group', False)

        project_group_vo = self.project_group_mgr.get_project_group(params['project_group_id'], domain_id)

        if release_parent_project_group:
            params['parent_project_group'] = None
            params['parent_project_group_id'] = None
        else:
            if 'parent_project_group_id' in params:
                params['parent_project_group'] = self._get_parent_project_group(
                    params['parent_project_group_id'], params['domain_id'], params['project_group_id'])

        return self.project_group_mgr.update_project_group_by_vo(params, project_group_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_group_id', 'domain_id'])
    def delete(self, params):
        """ Delete project group

        Args:
            params (dict): {
                'project_group_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        project_group_vo = self.project_group_mgr.get_project_group(params['project_group_id'], params['domain_id'])
        self.project_group_mgr.delete_project_group_by_vo(project_group_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_group_id', 'domain_id'])
    @change_only_key({'parent_project_group_info': 'parent_project_group'})
    def get(self, params):
        """ Get project group

        Args:
            params (dict): {
                'project_group_id': 'str',
                'domain_id': 'str',
                'only': 'list'
            }

        Returns:
            project_group_vo
        """

        return self.project_group_mgr.get_project_group(params['project_group_id'], params['domain_id'],
                                                        params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @change_only_key({'parent_project_group_info': 'parent_project_group'}, key_path='query.only')
    @append_query_filter(['project_group_id', 'name', 'parent_project_group_id', 'domain_id'])
    @append_keyword_filter(['project_group_id', 'name'])
    def list(self, params):
        """ List project groups

        Args:
            params (dict): {
                'project_group_id': 'str',
                'name': 'str',
                'parent_project_group_id': 'str',
                'author_within': 'bool',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of project_group_vo'
            total_count (int)
        """
        role_type = self.transaction.get_meta('authorization.role_type')
        user_projects = self.transaction.get_meta('authorization.projects')
        user_project_groups = self.transaction.get_meta('authorization.project_groups')
        author_within = params.get('author_within', False)
        query = params.get('query', {})
        domain_id = params['domain_id']

        if 'only' in query and 'parent_project_group' in query['only']:
            query['only'].append('parent_project_group_id')

        # For Access Control
        if role_type == 'PROJECT':
            if author_within:
                query['filter'].append({
                    'k': 'user_project_groups',
                    'v': user_project_groups,
                    'o': 'in'
                })
            else:
                self._append_user_project_group_filter(query, user_projects, user_project_groups, domain_id)

        project_group_vos, total_count = self.project_group_mgr.list_project_groups(query)

        return project_group_vos, total_count, self._get_parent_project_groups_info(project_group_vos)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id', 'user_project_groups'])
    @append_keyword_filter(['project_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_project_groups': 'list', # from meta
            }

        Returns:
            values (list): 'list of statistics data'
            total_count (int)
        """

        query = params.get('query', {})
        return self.project_group_mgr.stat_project_groups(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_group_id', 'user_id', 'role_id', 'domain_id'])
    def add_member(self, params):
        """ Add project group member

        Args:
            params (dict): {
                'project_group_id': 'str',
                'user_id': 'str',
                'is_external_user': 'bool',
                'role_id': 'str',
                'labels': 'list',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            role_binding_vo (object)
        """

        is_external_user = params.get('is_external_user', False)
        user_id = params['user_id']
        domain_id = params['domain_id']

        if is_external_user:
            self._create_external_user(user_id, domain_id)

        params['resource_type'] = 'identity.User'
        params['resource_id'] = params['user_id']
        del params['user_id']

        role_mgr: RoleManager = self.locator.get_manager('RoleManager')
        role_binding_mgr: RoleBindingManager = self.locator.get_manager('RoleBindingManager')

        role_vo = role_mgr.get_role(params['role_id'], params['domain_id'])
        if role_vo.role_type != 'PROJECT':
            raise ERROR_ONLY_PROJECT_ROLE_TYPE_ALLOWED()

        return role_binding_mgr.create_role_binding(params)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_group_id', 'user_id', 'domain_id'])
    def modify_member(self, params):
        """ Modify project group member

        Args:
            params (dict): {
                'project_group_id': 'str',
                'user_id': 'str',
                'labels': 'list',
                'tags': 'list',
                'domain_id': 'str'
            }

        Returns:
            role_binding_vo (object)
        """

        project_group_id = params['project_group_id']
        user_id = params['user_id']
        domain_id = params['domain_id']

        project_group_vo = self.project_group_mgr.get_project_group(project_group_id, domain_id)

        role_binding_mgr: RoleBindingManager = self.locator.get_manager('RoleBindingManager')
        role_binding_vos = role_binding_mgr.get_project_role_binding('identity.User', user_id, domain_id,
                                                                     project_group_vo=project_group_vo)

        if role_binding_vos.count() == 0:
            raise ERROR_NOT_FOUND(key='user_id', value=user_id)

        return role_binding_mgr.update_role_binding_by_vo(params, role_binding_vos[0])

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_group_id', 'domain_id', 'user_id'])
    def remove_member(self, params):
        """ Remove project group member

        Args:
            params (dict): {
                'project_group_id': 'str',
                'user_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        project_group_id = params['project_group_id']
        user_id = params['user_id']
        domain_id = params['domain_id']

        project_group_vo = self.project_group_mgr.get_project_group(project_group_id, domain_id)

        role_binding_mgr: RoleBindingManager = self.locator.get_manager('RoleBindingManager')
        role_binding_vos = role_binding_mgr.get_project_role_binding('identity.User', user_id, domain_id,
                                                                     project_group_vo=project_group_vo)

        if role_binding_vos.count() == 0:
            raise ERROR_NOT_FOUND(key='user_id', value=user_id)

        for role_binding_vo in role_binding_vos:
            role_binding_mgr.delete_role_binding_by_vo(role_binding_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_group_id', 'domain_id'])
    @change_only_key({'project_group_info': 'project_group', 'project_info': 'project', 'role_info': 'role'},
                     key_path='query.only')
    @append_query_filter(['user_id', 'role_id'])
    @append_keyword_filter(['resource_id'])
    def list_members(self, params):
        """ List users in project group

        Args:
            params (dict): {
                'project_group_id': 'str',
                'user_id': 'str',
                'role_id': 'str',
                'domain_id': 'str',
                'include_parent_member': 'bool',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of role_binding_vo'
            total_count (int)
        """

        project_group_id = params['project_group_id']
        domain_id = params['domain_id']
        include_parent_member = params.get('include_parent_member', False)

        role_binding_mgr: RoleBindingManager = self.locator.get_manager('RoleBindingManager')

        query = params.get('query', {})
        query['filter'] = query.get('filter', [])

        if include_parent_member:
            project_group_vo = self.project_group_mgr.get_project_group(project_group_id, domain_id)
            parent_project_group_vos = self._get_parents(project_group_vo.parent_project_group, [])
            query['filter'].append({
                'k': 'project_group',
                'v': parent_project_group_vos + [project_group_vo],
                'o': 'in'
            })
            query['filter'].append({
                'k': 'role_type',
                'v': 'PROJECT',
                'o': 'eq'
            })
        else:
            query['filter'].append({
                'k': 'project_group_id',
                'v': project_group_id,
                'o': 'eq'
            })

        return role_binding_mgr.list_role_bindings(query)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['project_group_id', 'domain_id'])
    @change_only_key({'project_group_info': 'project_group'}, key_path='query.only')
    @append_keyword_filter(['project_id', 'name'])
    def list_projects(self, params):
        """ List projects in project group

        Args:
            params (dict): {
                'project_group_id': 'str',
                'recursive': 'bool',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of project_vo'
            total_count (int)
        """

        role_type = self.transaction.get_meta('authorization.role_type')
        user_projects = self.transaction.get_meta('authorization.projects')
        project_group_id = params['project_group_id']
        domain_id = params['domain_id']
        recursive = params.get('recursive', False)
        query = params.get('query', {})

        if 'only' in query and 'project_group' in query['only']:
            query['only'].append('project_group_id')

        project_mgr: ProjectManager = self.locator.get_manager('ProjectManager')

        if 'filter' not in query:
            query['filter'] = []

        related_project_groups = []
        project_group_vo = self.project_group_mgr.get_project_group(project_group_id, domain_id)
        related_project_groups.append(project_group_vo)

        if recursive:
            related_project_groups = self._get_related_project_group(project_group_vo, related_project_groups)

        query['filter'].append({
            'k': 'project_group',
            'v': related_project_groups,
            'o': 'in'
        })

        if role_type == 'PROJECT':
            query['filter'].append({
                'k': 'user_projects',
                'v': user_projects,
                'o': 'in'
            })

        project_vos, total_count = project_mgr.list_projects(query)

        return project_vos, total_count, self._get_project_groups_info(project_vos)

    def _get_related_project_group(self, project_group_vo, related_project_groups):
        query = {
            'filter': [{
                'k': 'parent_project_group',
                'v': project_group_vo,
                'o': 'eq'
            }]
        }

        project_group_vos, total_count = self.project_group_mgr.list_project_groups(query)

        if total_count > 0:
            related_project_groups += project_group_vos
            for pg_vo in project_group_vos:
                related_project_groups = self._get_related_project_group(pg_vo, related_project_groups)

        return related_project_groups

    def _get_parent_project_group(self, parent_project_group_id, domain_id, project_group_id=None):
        query_filter = [{
            'k': 'project_group_id',
            'v': parent_project_group_id,
            'o': 'eq'
        }, {
            'k': 'domain_id',
            'v': domain_id,
            'o': 'eq'
        }]

        if project_group_id:
            query_filter.append({
                'k': 'project_group_id',
                'v': project_group_id,
                'o': 'not'
            })

        parent_project_group_vos, total_count = self.project_group_mgr.list_project_groups({'filter': query_filter})

        if total_count == 0:
            raise ERROR_NOT_FOUND(key='parent_project_group_id', value=parent_project_group_id)

        return parent_project_group_vos[0]

    def _append_user_project_group_filter(self, query, projects, project_groups, domain_id):
        self.project_mgr: ProjectManager = self.locator.get_manager('ProjectManager')
        query['filter'] = query.get('filter', [])
        if projects or project_groups:
            all_user_project_path = project_groups[:]
            for project_id in projects:
                if project_id is not None:
                    all_user_project_path += self.project_mgr.get_project_path(project_id, None, domain_id) or []

            for project_group_id in project_groups:
                if project_group_id is not None:
                    all_user_project_path += self.project_mgr.get_project_path(None, project_group_id, domain_id) or []

            query['filter'].append({
                'k': 'user_project_groups',
                'v': list(set(all_user_project_path)),
                'o': 'in'
            })

        return query

    def _get_parents(self, project_group_vo: ProjectGroup, parent_project_group_vos):
        parent_project_group_vos.append(project_group_vo)

        if project_group_vo.parent_project_group:
            return self._get_parents(project_group_vo.parent_project_group, parent_project_group_vos)
        else:
            return parent_project_group_vos

    def _create_external_user(self, user_id, domain_id):
        user_mgr: UserManager = self.locator.get_manager('UserManager')
        domain_mgr: DomainManager = self.locator.get_manager('DomainManager')

        user_vos = user_mgr.filter_users(user_id=user_id, domain_id=domain_id)

        if user_vos.count() == 0:
            domain_vo = domain_mgr.get_domain(domain_id)
            user_mgr.create_user({
                'user_id': user_id,
                'user_type': 'USER',
                'backend': 'EXTERNAL',
                'domain_id': domain_id
            }, domain_vo)

    def _get_project_groups_info(self, project_vos):
        project_groups_info = {}
        project_group_ids = []
        for project_vo in project_vos:
            if project_vo.project_group_id:
                project_group_ids.append(project_vo.project_group_id)

        pg_vos = self.project_group_mgr.filter_project_groups(project_group_id=project_group_ids)
        for pg_vo in pg_vos:
            project_groups_info[pg_vo.project_group_id] = pg_vo

        return project_groups_info

    def _get_parent_project_groups_info(self, project_group_vos):
        parent_project_groups_info = {}
        parent_project_group_ids = []
        for project_group_vo in project_group_vos:
            if project_group_vo.parent_project_group_id:
                parent_project_group_ids.append(project_group_vo.parent_project_group_id)

        if len(parent_project_group_ids) > 0:
            parent_pg_vos = self.project_group_mgr.filter_project_groups(project_group_id=parent_project_group_ids)
            for parent_pg_vo in parent_pg_vos:
                parent_project_groups_info[parent_pg_vo.project_group_id] = parent_pg_vo

        return parent_project_groups_info



from jsonschema import validate

from spaceone.core.service import *
from spaceone.identity.manager.service_account_manager import ServiceAccountManager
from spaceone.identity.manager.project_manager import ProjectManager
from spaceone.identity.manager.provider_manager import ProviderManager
from spaceone.identity.model.service_account_model import ServiceAccount
from spaceone.identity.error.error_service_account import *


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class ServiceAccountService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.service_account_mgr: ServiceAccountManager = self.locator.get_manager('ServiceAccountManager')

    @transaction(append_meta={
        'authorization.scope': 'DOMAIN_OR_PROJECT',
        'authorization.require_project_id': True
    })
    @check_required(['name', 'data', 'provider', 'service_account_type', 'domain_id'])
    def create(self, params):
        """
        Args:
            params (dict): {
                'name': 'str',
                'service_account_type': 'str',
                'data': 'dict',
                'provider': 'str',
                'trusted_service_account_id': 'str',
                'project_id': 'str',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            service_account_vo (object)
        """
        domain_id = params['domain_id']
        service_account_type = params['service_account_type']

        if 'project_id' in params and params['service_account_type'] == 'TRUSTED':
            raise ERROR_INVALID_PARAMETER(key='project_id',
                                          reason='Trusted service account cannot configure a project.')

        if service_account_type == 'TRUSTED':
            params.update({
                'trusted_service_account_id': None,
                'scope': 'DOMAIN'
            })
        elif service_account_type == 'GENERAL':
            params['scope'] = 'PROJECT'

            if trusted_service_account_id := params.get('trusted_service_account_id'):
                self._validation_trusted_service_account_check(trusted_service_account_id, domain_id)
        else:
            raise ERROR_INVALID_PARAMETER(key='service_account_type', reason=f'{service_account_type}')

        self._check_data(params['data'], params['provider'])

        if 'project_id' in params:
            params['project'] = self._get_project(params['project_id'], params['domain_id'])

        return self.service_account_mgr.create_service_account(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN_OR_PROJECT'})
    @check_required(['service_account_id', 'domain_id'])
    def update(self, params):
        """
        Args:
            params (dict): {
                'service_account_id': 'str',
                'name': 'str',
                'data': 'dict',
                'project_id': 'str',
                'tags': 'dict',
                'release_service_account': 'bool',
                'domain_id': 'str'
            }

        Returns:
            service_account_vo (object)
        """

        service_account_id = params['service_account_id']
        domain_id = params['domain_id']
        project_id = params.get('project_id')

        service_account_vo: ServiceAccount = self.service_account_mgr.get_service_account(service_account_id, domain_id)

        if 'data' in params:
            self._check_data(params['data'], service_account_vo.provider)

        if project_id:
            if service_account_vo.service_account_type == 'TRUSTED':
                raise ERROR_INVALID_PARAMETER(key='project_id',
                                              reason='Trusted service account cannot configure a project.')

            params['project'] = self._get_project(params['project_id'], params['domain_id'])

        if 'trusted_service_account_id' in params:
            self._validation_trusted_service_account_check(params['trusted_service_account_id'], domain_id)

        if params.get('release_trusted_service_account') is True:
            params['trusted_service_account_id'] = None

        service_account_vo = self.service_account_mgr.update_service_account_by_vo(params, service_account_vo)

        if project_id:
            self.service_account_mgr.update_secret_project(service_account_id, project_id, domain_id)

        return service_account_vo

    @transaction(append_meta={'authorization.scope': 'DOMAIN_OR_PROJECT'})
    @check_required(['service_account_id', 'domain_id'])
    def delete(self, params):
        """
        Args:
            params (dict): {
                'service_account_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        service_account_id = params['service_account_id']
        domain_id = params['domain_id']
        service_account_vo: ServiceAccount = self.service_account_mgr.get_service_account(service_account_id, domain_id)

        # self.service_account_mgr.check_service_account_secrets(service_account_id, domain_id,
        #                                                        service_account_vo.service_account_type)
        self.service_account_mgr.delete_service_account_secrets(service_account_id, domain_id,
                                                                service_account_vo.service_account_type)

        service_account_vo.delete()

    @transaction(append_meta={'authorization.scope': 'DOMAIN_OR_PROJECT'})
    @check_required(['service_account_id', 'domain_id'])
    @change_only_key({'project_info': 'project'})
    def get(self, params):
        """
        Args:
            params (dict): {
                'service_account_id': 'str',
                'domain_id': 'str',
                'only': 'list'
            }

        Returns:
            service_account_vo (object)
        """

        return self.service_account_mgr.get_service_account(params['service_account_id'], params['domain_id'],
                                                            params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN_OR_PROJECT'})
    @check_required(['domain_id'])
    @change_only_key({'project_info': 'project'}, key_path='query.only')
    @append_query_filter(['service_account_id', 'service_account_type', 'trusted_service_account_id',
                          'name', 'provider', 'scope', 'project_id', 'domain_id', 'user_projects'])
    @append_keyword_filter(['service_account_id', 'name', 'provider'])
    def list(self, params):
        """
        Args:
            params (dict): {
                    'service_account_id': 'str',
                    'service_account_type': 'str',
                    'trusted_service_account_id': 'str',
                    'name': 'str',
                    'scope': 'str',
                    'provider': 'str',
                    'project_id': 'str',
                    'has_secret': 'bool',
                    'domain_id': 'str',
                    'query': 'dict (spaceone.api.core.v1.Query)',
                    'user_projects': 'list', // from meta
                }

        Returns:
            results (list): 'list of service_account_vo'
            total_count (int)
        """

        query = params.get('query', {})

        has_secret = params.get('has_secret', False)
        if has_secret:
            query = self._append_secret_filter(query, params['domain_id'])

        service_account_vos, total_count = self.service_account_mgr.list_service_accounts(query)

        return service_account_vos, total_count, self._get_project_info(service_account_vos)

    @transaction(append_meta={'authorization.scope': 'DOMAIN_OR_PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['project_id', 'domain_id', 'user_projects'])
    @append_keyword_filter(['service_account_id', 'name', 'provider'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'has_secret': 'bool',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta
            }

        Returns:
            values (list): 'list of statistics data'
            total_count (int)
        """

        query = params.get('query', {})

        has_secret = params.get('has_secret', False)
        if has_secret:
            query = self._append_secret_filter(query, params['domain_id'])

        return self.service_account_mgr.stat_service_accounts(query)

    def _check_data(self, data, provider):
        provider_mgr: ProviderManager = self.locator.get_manager('ProviderManager')
        provider_vo = provider_mgr.get_provider(provider)
        schema = provider_vo.template.get('service_account', {}).get('schema')

        if schema:
            try:
                validate(instance=data, schema=schema)
            except Exception as e:
                raise ERROR_INVALID_PARAMETER(key='data', reason=e.message)
        else:
            if data != {}:
                raise ERROR_INVALID_PARAMETER(key='data', reason='data format is invalid.')

    def _validation_trusted_service_account_check(self, trusted_service_account_id, domain_id):
        query = {
            'filter': [
                {'k': 'service_account_id', 'v': trusted_service_account_id, 'o': 'eq'},
                {'k': 'service_account_type', 'v': 'TRUSTED', 'o': 'eq'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'},
            ]
        }
        results, total_count = self.service_account_mgr.list_service_accounts(query)
        if total_count == 0:
            raise ERROR_NOT_FOUND_TRUSTED_SERVICE_ACCOUNT_ID(trusted_service_account_id=trusted_service_account_id)

    def _get_project(self, project_id, domain_id):
        project_mgr: ProjectManager = self.locator.get_manager('ProjectManager')
        return project_mgr.get_project(project_id, domain_id)

    def _get_project_info(self, service_account_vos):
        project_mgr: ProjectManager = self.locator.get_manager('ProjectManager')

        projects_info = {}
        project_ids = []
        for service_account_vo in service_account_vos:
            if service_account_vo.project_id:
                project_ids.append(service_account_vo.project_id)

        project_vos = project_mgr.filter_projects(project_id=project_ids)
        for project_vo in project_vos:
            projects_info[project_vo.project_id] = project_vo

        return projects_info

    def _append_secret_filter(self, query, domain_id):
        service_account_ids = self.service_account_mgr.get_all_service_account_ids_using_secret(domain_id)
        query['filter'] = query.get('filter', [])
        query['filter'].append({
            'k': 'service_account_id',
            'v': service_account_ids,
            'o': 'in'
        })

        return query



from spaceone.core.service import *
from spaceone.core import utils
from spaceone.identity.manager import PolicyManager


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PolicyService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.policy_mgr: PolicyManager = self.locator.get_manager('PolicyManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['name', 'permissions', 'domain_id'])
    def create(self, params):
        """ Create policy

        Args:
            params (dict): {
                'name': 'str',
                'permissions': 'list',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            policy_vo (object)
        """

        return self.policy_mgr.create_policy(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['policy_id', 'domain_id'])
    def update(self, params):
        """ Update policy

        Args:
            params (dict): {
                'policy_id': 'str',
                'name': 'str',
                'permissions': 'list',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            policy_vo (object)
        """

        return self.policy_mgr.update_policy(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['policy_id', 'domain_id'])
    def delete(self, params):
        """ Delete policy

        Args:
            params (dict): {
                'policy_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.policy_mgr.delete_policy(params['policy_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['policy_id', 'domain_id'])
    def get(self, params):
        """ Get policy

        Args:
            params (dict): {
                'policy_id': 'str',
                'domain_id': 'str',
                'only': 'list'
            }

        Returns:
            domain_vo (object)
        """

        return self.policy_mgr.get_policy(params['policy_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @append_query_filter(['policy_id', 'name', 'domain_id'])
    @append_keyword_filter(['policy_id', 'name'])
    def list(self, params):
        """ List polices

        Args:
            params (dict): {
                'policy_id': 'str',
                'name': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of policy_vo'
            total_count (int)
        """

        query = self._append_policy_type_filter(params.get('query', {}))
        return self.policy_mgr.list_policies(query)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @append_keyword_filter(['policy_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list): 'list of statistics data'
            total_count (int)
        """

        query = self._append_policy_type_filter(params.get('query', {}))
        return self.policy_mgr.stat_policies(query)

    @staticmethod
    def _append_policy_type_filter(query):
        query['filter'] = query.get('filter', [])
        query['filter'].append({
            'k': 'policy_type',
            'v': 'CUSTOM',
            'o': 'eq'
        })

        return query



import logging

from spaceone.core import cache
from spaceone.core.auth.jwt import JWTAuthenticator, JWTUtil
from spaceone.core.service import *
from spaceone.identity.error.error_authentication import *
from spaceone.identity.error.error_domain import *
from spaceone.identity.manager import DomainManager, DomainSecretManager, UserManager
from spaceone.identity.model import User, Domain

_LOGGER = logging.getLogger(__name__)


@event_handler
class TokenService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.user_mgr: UserManager = self.locator.get_manager('UserManager')
        self.domain_mgr: DomainManager = self.locator.get_manager('DomainManager')
        self.domain_secret_mgr: DomainSecretManager = self.locator.get_manager('DomainSecretManager')

    @transaction
    @check_required(['credentials', 'domain_id'])
    def issue(self, params):
        """ Issue token

        Args:
            params (dict): {
                'user_id': 'str',
                'credentials': 'dict'
                'user_type': 'str',
                'timeout': 'int',
                'refresh_count': 'int',
                'domain_id': 'str'
            }

        Returns:
            result (dict): {
                'access_token': 'str',
                'refresh_token': 'str'
            }
        """

        user_id = params.get('user_id')
        user_type = params.get('user_type', 'USER')
        domain_id = params['domain_id']
        timeout = params.get('timeout')
        refresh_count = params.get('refresh_count')

        private_jwk = self.domain_secret_mgr.get_domain_private_key(domain_id=domain_id)
        refresh_private_jwk = self.domain_secret_mgr.get_domain_refresh_private_key(domain_id=domain_id)

        token_manager = self._get_token_manager(user_id, user_type, domain_id)
        token_manager.authenticate(user_id, domain_id, params['credentials'])

        token_info = token_manager.issue_token(private_jwk=private_jwk, refresh_private_jwk=refresh_private_jwk,
                                               timeout=timeout, ttl=refresh_count)

        return token_info

    @transaction
    def refresh(self, params):
        """ Refresh token

        Args:
            params (dict): {}

        Returns:
            result (dict): {
                'access_token': 'str',
                'refresh_token': 'str'
            }
        """

        refresh_token = self.transaction.get_meta('token')

        if refresh_token is None:
            raise ERROR_INVALID_REFRESH_TOKEN()

        domain_id = self._extract_domain_id(refresh_token)

        private_jwk = self.domain_secret_mgr.get_domain_private_key(domain_id=domain_id)
        refresh_public_jwk = self.domain_secret_mgr.get_domain_refresh_public_key(domain_id=domain_id)
        refresh_private_jwk = self.domain_secret_mgr.get_domain_refresh_private_key(domain_id=domain_id)

        token_info = self._verify_refresh_token(refresh_token, refresh_public_jwk)
        token_mgr = self._get_token_manager(token_info['user_id'], token_info['user_type'], domain_id)
        token_mgr.check_refreshable(token_info['key'], token_info['ttl'])

        return token_mgr.refresh_token(token_info['user_id'], domain_id, ttl=token_info['ttl']-1,
                                       private_jwk=private_jwk, refresh_private_jwk=refresh_private_jwk)

    def _get_token_manager(self, user_id, user_type, domain_id):
        self._check_domain_state(domain_id)

        # user_type = USER | DOMAIN_OWNER
        if user_type == 'DOMAIN_OWNER':
            return self.locator.get_manager('DomainOwnerTokenManager')

        if user_id:
            user_backend = self._get_user_backend(user_id, domain_id)

            if user_backend == 'LOCAL':
                return self.locator.get_manager('LocalTokenManager')
            else:
                return self.locator.get_manager('ExternalTokenManager')
        else:
            return self.locator.get_manager('ExternalTokenManager')

    @cache.cacheable(key='user-backend:{domain_id}:{user_id}', expire=600)
    def _get_user_backend(self, user_id, domain_id):
        try:
            user_vo: User = self.user_mgr.get_user(user_id, domain_id)
        except Exception as e:
            _LOGGER.error(f'[_get_user_backend] Authentication failure: {getattr(e, "message", e)}')
            raise ERROR_AUTHENTICATION_FAILURE(user_id=user_id)

        return user_vo.backend

    @cache.cacheable(key='domain-state:{domain_id}', expire=3600)
    def _check_domain_state(self, domain_id):
        domain_vo: Domain = self.domain_mgr.get_domain(domain_id)

        if domain_vo.state != 'ENABLED':
            raise ERROR_DOMAIN_STATE(domain_id=domain_vo.domain_id)

    @staticmethod
    def _extract_domain_id(token):
        try:
            decoded = JWTUtil.unverified_decode(token)
        except Exception as e:
            _LOGGER.error(f'[_extract_domain_id] {e}')
            _LOGGER.error(token)
            raise ERROR_AUTHENTICATE_FAILURE(message='Cannot decode token.')

        domain_id = decoded.get('did')

        if domain_id is None:
            raise ERROR_AUTHENTICATE_FAILURE(message='Empty domain_id provided.')

        return domain_id

    @staticmethod
    def _verify_refresh_token(token, public_jwk):
        try:
            decoded = JWTAuthenticator(public_jwk).validate(token)
        except Exception as e:
            _LOGGER.error(f'[_verify_refresh_token] {e}')
            raise ERROR_AUTHENTICATE_FAILURE(message='Token validation failed.')

        if decoded.get('cat') != 'REFRESH_TOKEN':
            raise ERROR_INVALID_REFRESH_TOKEN()

        return {
            'user_id': decoded['aud'],
            'user_type': decoded['user_type'],
            'key': decoded['key'],
            'ttl': decoded['ttl'],
            'iat': decoded['iat'],
            'exp': decoded['exp']
        }



from mongoengine import *
from spaceone.core.error import *
from spaceone.core.model.mongo_model import MongoModel
from spaceone.identity.model.role_model import Role
from spaceone.identity.model.project_model import Project
from spaceone.identity.model.project_group_model import ProjectGroup
from spaceone.identity.model.user_model import User


class RoleBinding(MongoModel):
    role_binding_id = StringField(max_length=40, generate_id='rb', unique=True)
    resource_type = StringField(max_length=255)
    resource_id = StringField(max_length=255)
    role = ReferenceField('Role', reverse_delete_rule=DENY)
    project = ReferenceField('Project', null=True, default=None, reverse_delete_rule=CASCADE)
    project_group = ReferenceField('ProjectGroup', null=True, default=None, reverse_delete_rule=CASCADE)
    user = ReferenceField('User', null=True, default=None, reverse_delete_rule=CASCADE)
    role_id = StringField(max_length=40)
    project_id = StringField(max_length=40, null=True, default=None)
    project_group_id = StringField(max_length=40, null=True, default=None)
    labels = ListField(StringField(max_length=255))
    tags = DictField()
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)

    meta = {
        'updatable_fields': [
            'labels',
            'tags',
            'role_id',
            'project_id',
            'project_group_id'
        ],
        'minimal_fields': [
            'role_binding_id',
            'resource_type',
            'resource_id',
            'role'
        ],
        'change_query_keys': {
            'user_id': 'user.user_id',
            'name': 'user.name',
            'email': 'user.email',
            'role_type': 'role.role_type',
            'project_group_id': 'project_group.project_group_id'
        },
        'reference_query_keys': {
            'user': User,
            'role': Role,
            'project': Project,
            'project_group': ProjectGroup
        },
        'ordering': ['resource_type', 'resource_id'],
        'indexes': [
            # 'role_binding_id',
            'resource_type',
            'resource_id',
            'role',
            'project',
            'project_group',
            'role_id',
            'project_id',
            'project_group_id',
            'domain_id',
        ]
    }



from mongoengine import *
from spaceone.core.error import *
from spaceone.core.model.mongo_model import MongoModel


class RolePolicy(EmbeddedDocument):
    policy_type = StringField(max_length=20, choices=('MANAGED', 'CUSTOM'))
    policy = ReferenceField('Policy')
    policy_id = StringField(max_length=40)

    def to_dict(self):
        return dict(self.to_mongo())


class PagePermission(EmbeddedDocument):
    page = StringField(max_length=255, required=True)
    permission = StringField(max_length=20, choices=('VIEW', 'MANAGE'))

    def to_dict(self):
        return dict(self.to_mongo())


class Role(MongoModel):
    role_id = StringField(max_length=40, generate_id='role', unique=True)
    name = StringField(max_length=255, unique_with='domain_id')
    role_type = StringField(max_length=20, choices=('SYSTEM', 'DOMAIN', 'PROJECT'))
    tags = DictField(default={})
    policies = ListField(EmbeddedDocumentField(RolePolicy))
    page_permissions = ListField(EmbeddedDocumentField(PagePermission), default=[])
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)

    meta = {
        'updatable_fields': [
            'name',
            'policies',
            'page_permissions',
            'tags'
        ],
        'minimal_fields': [
            'role_id',
            'name',
            'role_type',
            'page_permissions'
        ],
        'change_query_keys': {
            'policy_id': 'policies.policy_id'
        },
        'ordering': ['name'],
        'indexes': [
            # 'role_id',
            'role_type',
            'domain_id',
            'tags'
        ]
    }



import logging

from spaceone.core.manager import BaseManager
from spaceone.identity.conf.provider_conf import DEFAULT_PROVIDERS
from spaceone.identity.model.provider_model import Provider

_LOGGER = logging.getLogger(__name__)


class ProviderManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.provider_model: Provider = self.locator.get_model('Provider')

    def create_provider(self, params):
        def _rollback(provider_vo):
            _LOGGER.info(f'[create_provider._rollback] Create provider : {provider_vo.provider}')
            provider_vo.delete()

        provider_vo: Provider = self.provider_model.create(params)
        self.transaction.add_rollback(_rollback, provider_vo)

        return provider_vo

    def update_provider(self, params):
        def _rollback(old_data):
            _LOGGER.info(f'[update_provider._rollback] Revert Data : {old_data["provider"]}')
            provider_vo.update(old_data)

        provider_vo: Provider = self.get_provider(params['provider'])
        self.transaction.add_rollback(_rollback, provider_vo.to_dict())

        return provider_vo.update(params)

    def delete_provider(self, provider):
        provider_vo: Provider = self.get_provider(provider)
        provider_vo.delete()

    def get_provider(self, provider, only=None):
        return self.provider_model.get(provider=provider, only=only)

    def list_providers(self, query={}):
        return self.provider_model.query(**query)

    def stat_providers(self, query):
        return self.provider_model.stat(**query)

    def create_default_providers(self, installed_providers):
        for provider in DEFAULT_PROVIDERS:
            if provider['provider'] not in installed_providers:
                _LOGGER.debug(f'Create default provider: {provider["name"]}')
                self.create_provider(provider)



from spaceone.identity.info.api_key_info import *
from spaceone.identity.info.authorization_info import *
from spaceone.identity.info.common_info import *
from spaceone.identity.info.domain_info import *
from spaceone.identity.info.policy_info import *
from spaceone.identity.info.project_group_info import *
from spaceone.identity.info.project_info import *
from spaceone.identity.info.role_info import *
from spaceone.identity.info.role_binding_info import *
from spaceone.identity.info.token_info import *
from spaceone.identity.info.user_info import *
from spaceone.identity.info.domain_owner_info import *
from spaceone.identity.info.find_user_info import *
from spaceone.identity.info.provider_info import *
from spaceone.identity.info.service_account_info import *
from spaceone.identity.info.endpoint_info import *



from spaceone.identity.model.api_key_model import APIKey
from spaceone.identity.model.domain_model import Domain
from spaceone.identity.model.domain_owner_model import DomainOwner
from spaceone.identity.model.project_group_model import ProjectGroup
from spaceone.identity.model.project_model import Project
from spaceone.identity.model.provider_model import Provider
from spaceone.identity.model.service_account_model import ServiceAccount
from spaceone.identity.model.domain_secret_model import DomainSecret
from spaceone.identity.model.policy_model import Policy
from spaceone.identity.model.role_model import Role
from spaceone.identity.model.role_binding_model import RoleBinding
from spaceone.identity.model.user_model import User



import functools
from spaceone.api.identity.v1 import endpoint_pb2

__all__ = ['EndpointInfo', 'EndpointsInfo']


def EndpointInfo(endpoint_vo: dict, minimal=False):
    info = {
        'service': endpoint_vo['service'],
        'name': endpoint_vo['name'],
        'endpoint': endpoint_vo['endpoint'],
    }

    if not minimal:
        info.update({
            'state': endpoint_vo.get('state'),
            'version': endpoint_vo.get('version'),
        })

    return endpoint_pb2.EndpointInfo(**info)


def EndpointsInfo(endpoint_vos, total_count, **kwargs):
    results = list(map(functools.partial(EndpointInfo, **kwargs), endpoint_vos))

    return endpoint_pb2.EndpointsInfo(results=results, total_count=total_count)



import logging
from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.identity.model.policy_model import *
from spaceone.identity.model.role_model import *

_LOGGER = logging.getLogger(__name__)


class RoleManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.role_model: Role = self.locator.get_model('Role')

    def create_role(self, params):
        def _rollback(role_vo):
            _LOGGER.info(f'[create_role._rollback] Delete role : {role_vo.name} ({role_vo.role_id})')
            role_vo.delete()

        role_vo = self.role_model.create(params)
        self.transaction.add_rollback(_rollback, role_vo)

        return role_vo

    def update_role(self, params):
        def _rollback(old_data):
            _LOGGER.info(f'[update_role._rollback] Revert Data : {old_data["name"], ({old_data["role_id"]})}')
            role_vo.update(old_data)

        role_vo: Role = self.get_role(params['role_id'], params['domain_id'])
        self.transaction.add_rollback(_rollback, role_vo.to_dict())

        cache.delete_pattern(f'role-permissions:*{params["role_id"]}')
        cache.delete_pattern(f'user-permissions:*{params["role_id"]}*')

        return role_vo.update(params)

    def delete_role(self, role_id, domain_id):
        role_vo = self.get_role(role_id, domain_id)
        role_vo.delete()

        cache.delete_pattern(f'role-permissions:*{role_id}')
        cache.delete_pattern(f'user-permissions:*{role_id}*')

    def get_role(self, role_id, domain_id, only=None):
        return self.role_model.get(role_id=role_id, domain_id=domain_id, only=only)

    def filter_roles(self, **conditions):
        return self.role_model.filter(**conditions)

    def list_roles(self, query):
        return self.role_model.query(**query)

    def stat_roles(self, query):
        return self.role_model.stat(**query)



import functools
from spaceone.api.identity.v1 import api_key_pb2
from spaceone.core import utils
from spaceone.identity.model.api_key_model import APIKey

__all__ = ['APIKeyInfo', 'APIKeysInfo']


def APIKeyInfo(api_key_vo: APIKey, minimal=False, api_key=None, **kwargs):
    info = {
        'api_key': api_key,
        'api_key_id': api_key_vo.api_key_id,
        'state': api_key_vo.state,
        'user_id': api_key_vo.user_id
    }

    if not minimal:
        info.update({
            'domain_id': api_key_vo.domain_id,
            'last_accessed_at': utils.datetime_to_iso8601(api_key_vo.last_accessed_at),
            'created_at': utils.datetime_to_iso8601(api_key_vo.created_at)
        })

    return api_key_pb2.APIKeyInfo(**info)


def APIKeysInfo(api_key_vos, total_count, **kwargs):
    results = list(map(functools.partial(APIKeyInfo, **kwargs), api_key_vos))
    return api_key_pb2.APIKeysInfo(results=results, total_count=total_count)



import functools
from spaceone.api.identity.v1 import role_binding_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.identity.model.role_binding_model import RoleBinding
from spaceone.identity.info.role_info import RoleInfo
from spaceone.identity.info.project_info import ProjectInfo
from spaceone.identity.info.project_group_info import ProjectGroupInfo

__all__ = ['RoleBindingInfo', 'RoleBindingsInfo']


def RoleBindingInfo(role_binding_vo: RoleBinding, minimal=False):
    info = {
        'role_binding_id': role_binding_vo.role_binding_id,
        'resource_type': role_binding_vo.resource_type,
        'resource_id': role_binding_vo.resource_id,
        'role_info': RoleInfo(role_binding_vo.role, minimal=True) if role_binding_vo.role else None
    }

    if not minimal:
        info.update({
            'project_info': ProjectInfo(role_binding_vo.project, minimal=True) if role_binding_vo.project else None,
            'project_group_info': ProjectGroupInfo(role_binding_vo.project_group, minimal=True) if role_binding_vo.project_group else None,
            'labels': change_list_value_type(role_binding_vo.labels),
            'tags': change_struct_type(role_binding_vo.tags),
            'domain_id': role_binding_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(role_binding_vo.created_at)
        })

        if not role_binding_vo.project_id and role_binding_vo.project:
            role_binding_vo.update({'project_id': role_binding_vo.project.project_id})

        if not role_binding_vo.project_group_id and role_binding_vo.project_group:
            role_binding_vo.update({'project_group_id': role_binding_vo.project_group.project_group_id})

    return role_binding_pb2.RoleBindingInfo(**info)


def RoleBindingsInfo(role_binding_vos, total_count, **kwargs):
    results = list(map(functools.partial(RoleBindingInfo, **kwargs), role_binding_vos))

    return role_binding_pb2.RoleBindingsInfo(results=results, total_count=total_count)



from spaceone.api.identity.v1 import service_account_pb2, service_account_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class ServiceAccount(BaseAPI, service_account_pb2_grpc.ServiceAccountServicer):

    pb2 = service_account_pb2
    pb2_grpc = service_account_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ServiceAccountService', metadata) as service_account_svc:
            data = service_account_svc.create(params)
            return self.locator.get_info('ServiceAccountInfo', data)

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ServiceAccountService', metadata) as service_account_svc:
            data = service_account_svc.update(params)
            return self.locator.get_info('ServiceAccountInfo', data)

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ServiceAccountService', metadata) as service_account_svc:
            service_account_svc.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ServiceAccountService', metadata) as service_account_svc:
            data = service_account_svc.get(params)
            return self.locator.get_info('ServiceAccountInfo', data)

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ServiceAccountService', metadata) as service_account_svc:
            service_account_vos, total_count, projects_info = service_account_svc.list(params)
            return self.locator.get_info('ServiceAccountsInfo', service_account_vos, total_count,
                                         minimal=self.get_minimal(params), projects_info=projects_info)

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ServiceAccountService', metadata) as service_account_svc:
            return self.locator.get_info('StatisticsInfo', service_account_svc.stat(params))



import logging
import time
from typing import Tuple, Any

from spaceone.core import utils
from spaceone.core.auth.jwt.jwt_util import JWTUtil
from spaceone.identity.error import ERROR_GENERATE_KEY_FAILURE

_LOGGER = logging.getLogger(__name__)


class KeyGenerator:

    def __init__(self, prv_jwk, domain_id, audience):
        self.prv_jwk = prv_jwk
        self.domain_id = domain_id
        self.audience = audience

        self._do_parameter_check()

    def _do_parameter_check(self):
        if self.prv_jwk is None:
            raise ERROR_GENERATE_KEY_FAILURE()
        if self.domain_id is None:
            raise ERROR_GENERATE_KEY_FAILURE()

    def generate_api_key(self, api_key_id: str, user_type: str = 'USER') -> Tuple[str, Any]:
        payload = {
            'cat': 'API_KEY',
            'user_type': user_type,
            'did': self.domain_id,
            'aud': self.audience,
            'iat': int(time.time()),
            'api_key_id': api_key_id,
            'ver': '2020-12-07'
        }

        encoded = JWTUtil.encode(payload, self.prv_jwk)

        _LOGGER.debug(f'[KeyGenerator] Generated payload. ( '
                      f'cat: {payload.get("cat")}, '
                      f'user_type: {payload.get("user_type")}, '
                      f'did: {payload.get("did")}, '
                      f'aud: {payload.get("aud")}, '
                      f'api_key_id: {payload.get("api_key_id")}, '
                      f'version: {payload.get("version")} )')

        return encoded



from mongoengine import *

from spaceone.core.error import *
from spaceone.core.model.mongo_model import MongoModel


class DomainOwner(MongoModel):
    owner_id = StringField(max_length=40, unique_with='domain_id', required=True)
    password = BinaryField()
    name = StringField(max_length=128)
    email = StringField(max_length=255, default=None, null=True)
    language = StringField(max_length=7, default='en')
    timezone = StringField(max_length=50, default='UTC')
    domain_id = StringField(max_length=40)
    last_accessed_at = DateTimeField(default=None, null=True)
    created_at = DateTimeField(auto_now_add=True)

    meta = {
        'updatable_fields': [
            'password',
            'name',
            'email',
            'language',
            'timezone',
            'last_accessed_at'
        ],
        'minimal_fields': [
            'owner_id',
            'name'
        ],
        'ordering': ['name'],
        'indexes': [
            'owner_id',
            'email',
            'domain_id',
            'last_accessed_at'
        ]
    }

    @classmethod
    def create(cls, data):
        user_vos = cls.filter(domain_id=data['domain_id'])
        if user_vos.count() > 0:
            raise ERROR_NOT_UNIQUE(key='owner_id', value=data['owner_id'])
        return super().create(data)

    def update(self, data):
        return super().update(data)



from spaceone.api.identity.v1 import project_group_pb2, project_group_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class ProjectGroup(BaseAPI, project_group_pb2_grpc.ProjectGroupServicer):

    pb2 = project_group_pb2
    pb2_grpc = project_group_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            return self.locator.get_info('ProjectGroupInfo', project_group_svc.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            return self.locator.get_info('ProjectGroupInfo', project_group_svc.update(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            project_group_svc.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            return self.locator.get_info('ProjectGroupInfo', project_group_svc.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            project_group_vos, total_count, parent_project_groups_info = project_group_svc.list(params)
            return self.locator.get_info('ProjectGroupsInfo', project_group_vos, total_count,
                                         minimal=self.get_minimal(params),
                                         parent_project_groups_info=parent_project_groups_info)

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            return self.locator.get_info('StatisticsInfo', project_group_svc.stat(params))

    def add_member(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            return self.locator.get_info('ProjectGroupRoleBindingInfo', project_group_svc.add_member(params))

    def modify_member(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            return self.locator.get_info('ProjectGroupRoleBindingInfo', project_group_svc.modify_member(params))

    def remove_member(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            project_group_svc.remove_member(params)
            return self.locator.get_info('EmptyInfo')

    def list_members(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            project_group_map_vos, total_count = project_group_svc.list_members(params)
            return self.locator.get_info('ProjectGroupRoleBindingsInfo', project_group_map_vos, total_count,
                                         minimal=self.get_minimal(params))

    def list_projects(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProjectGroupService', metadata) as project_group_svc:
            project_vos, total_count, project_groups_info = project_group_svc.list_projects(params)
            return self.locator.get_info('ProjectGroupProjectsInfo', project_vos, total_count,
                                         minimal=self.get_minimal(params), project_groups_info=project_groups_info)



from spaceone.api.identity.v1 import user_pb2, user_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class User(BaseAPI, user_pb2_grpc.UserServicer):

    pb2 = user_pb2
    pb2_grpc = user_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            return self.locator.get_info('UserInfo', user_svc.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            return self.locator.get_info('UserInfo', user_svc.update(params))

    def verify_email(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            user_svc.verify_email(params)
            return self.locator.get_info('EmptyInfo')

    def confirm_email(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            return self.locator.get_info('UserInfo', user_svc.confirm_email(params))

    def reset_password(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            user_svc.reset_password(params)
            return self.locator.get_info('EmptyInfo')

    def set_required_actions(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            return self.locator.get_info('UserInfo', user_svc.set_required_actions(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            user_svc.delete(params)
            return self.locator.get_info('EmptyInfo')

    def enable(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            return self.locator.get_info('UserInfo', user_svc.enable(params))

    def disable(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            return self.locator.get_info('UserInfo', user_svc.disable(params))

    def find(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            users, total_count = user_svc.find(params)
            return self.locator.get_info('FindUsersInfo', users, total_count)

    def sync(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            pass
            # data = user_service.find_user(params)
            # return self.locator.get_info('UserInfo', data)

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            return self.locator.get_info('UserInfo', user_svc.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            users, total_count = user_svc.list(params)
            return self.locator.get_info('UsersInfo', users, total_count,
                                         minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('UserService', metadata) as user_svc:
            return self.locator.get_info('StatisticsInfo', user_svc.stat(params))
